import xarray as xr
from xarray import DataArray, Dataset
import uxarray as uxr
from uxarray import UxDataArray, UxDataset
from glob import glob
import os
import numpy as np
from scipy.optimize import curve_fit, OptimizeWarning
import shutil
import tempfile
from abc import ABC, abstractmethod
from typing import Tuple, List, Literal, get_args, Any, Union
from typing_extensions import Type
import hashlib
from numpy.typing import NDArray, ArrayLike
from numpy.random import Generator
from .target import Target
from ..utils.general_utils import validate_and_convert_location
from ..utils.custom_types import FloatLike, PairOfFloats
from ..utils.montecarlo import get_random_location_on_face
import warnings

# Define valid grid types
GridType = Literal["icosphere","arbitrary", "hires local"]

# Derive valid_grid_types list from GridType
valid_grid_types: Type[List[str]] = list(get_args(GridType))

# Default file names and directories
_DATA_DIR = "surface_data"
_COMBINED_DATA_FILE_NAME = "surf.nc"
_GRID_FILE_NAME = "grid.nc"

# This is a factor used to determine the smallest length scale in the grid
_SMALLFAC = 1.0e-5

class GridStrategy(ABC):
    def __init__(self, **kwargs: Any):
        self._grid = None

    @abstractmethod
    def generate_face_distribution(self) -> Tuple[NDArray,NDArray,NDArray]:
        pass
    

    def generate_grid(self,
                      grid_file: os.PathLike,
                      grid_hash: str | None = None,
                      **kwargs: Any) -> Tuple[os.PathLike, os.PathLike]:                       
        """
        Generate a tessellated mesh of a sphere of evenly distributed points


        Parameters
        ----------
        grid_file : os.PathLike
            The file path to the grid file.
        grid_hash : str, optional
            Hash of the grid parameters. Default is None, which will generate a new hash.
            
        Notes
        -----
        The grid configuration is determined by the `grid_strategy` attribute of the Surface object. The `grid_strategy` attribute
        determines the type of grid to be generated and its associated parameters. For detailed information on the parameters specific
        to each grid type, refer to the documentation of the respective grid parameter classes (`UnifromIcosphereGrid`, 
        `ArbitraryResolutionGrid`, `HiResLocalGrid`, etc.).
        
        """       

        points = self.generate_face_distribution() 
        grid = uxr.Grid.from_points(points, method="spherical_voronoi")
        if not grid_hash:
            grid_hash = self.generate_hash() 
        grid.attrs["grid_hash"] = grid_hash
        with tempfile.NamedTemporaryFile(delete=False) as temp_file:
            grid.to_xarray().to_netcdf(temp_file.name)
            temp_file.flush()
            os.fsync(temp_file.fileno())
            
        # Replace the original file only if writing succeeded
        shutil.move(temp_file.name,grid_file)            
        print("Mesh generation complete")
        self.grid = grid 
        return         
    
                         
    def generate_hash(self) -> str:
        """
        Generate a hash of the grid parameters.

        Returns
        -------
        str
            Hash of the grid parameters.
        """
        attribute_pairs = []
        for attr, value in vars(self).items():
            attribute_pairs.append(f"{attr}:{value}")
        combined = ":".join(attribute_pairs) 
        hash_object = hashlib.sha256(combined.encode())
        return hash_object.hexdigest()


    def check_if_regrid(self,
                         grid_file: os.PathLike,
                         **kwargs: Any,
                        ) -> bool:
        """
        Check if the existing grid matches the desired parameters determine if regridding is necessary.

        This function checks if a grid file exists and matches the specified parameters based on a unique hash generated from these 
        parameters. If the grid does not exist or does not match the parameters it returns True. 

        Parameters
        ----------
        grid_file : PathLike
            The file path where the grid is saved or will be saved. 
        grid_file : os.PathLike
            The file path to the grid file. 

        Returns
        -------
        bool
            A boolean indicating whether the grid should be regenerated. 
        """
    
        # Generate the hash for the current parameters
        grid_hash = self.generate_hash()

        # Find out if the file exists, if it does't we'll need to make a new grid
        make_new_grid = not os.path.exists(grid_file)
        
        if not make_new_grid:
            uxgrid = uxr.open_grid(grid_file)
            try: 
                old_hash = uxgrid.attrs.get("grid_hash")
                make_new_grid = old_hash != grid_hash
            except:
                make_new_grid = True
                
        return make_new_grid

    
    def create_grid(self,
                     grid_file: os.PathLike,
                     **kwargs: Any,
                     ) -> bool:
        """

        Creates a new grid file based on the grid parameters and stores the new grid as the grid property of the object. 

        Parameters
        ----------
        grid_file : PathLike
            The file path where the grid will be saved. 
        grid_file : os.PathLike
            The file path to the grid file. 
        """
    
        # Generate the hash for the current parameters
        grid_hash = self.generate_hash()

        if os.path.exists(grid_file):
            os.remove(grid_file)
        self.generate_grid(grid_file=grid_file, grid_hash=grid_hash, **kwargs) 
        
        # Check to make sure we can open the grid file, then store the hash in the metadata
        uxgrid = uxr.open_grid(grid_file)
        new_hash = uxgrid.attrs.get("grid_hash")
        assert(new_hash == grid_hash)

        return 
    
    @property
    def pix(self):
        """
        The approximate face size for the mesh inside the local region in meters.
        """
        return self._pix
    
    @pix.setter
    def pix(self, value: FloatLike):
        if not isinstance(value, FloatLike) or np.isnan(value) or np.isinf(value) or value <= 0:
            raise TypeError("pix must be a positive float")
        self._pix = value
        
    @property
    def radius(self):
        """
        The radius of the target body in meters.
        """
        return self._radius
    
    @radius.setter
    def radius(self, value: FloatLike):
        if not isinstance(value, FloatLike) or np.isnan(value) or np.isinf(value) or value <= 0:
            raise TypeError("radius must be a positive float")
        self._radius = value
        
    @property
    def grid(self):
        """
        The grid object.
        """
        return self._grid
    
    @grid.setter
    def grid(self, value: uxr.Grid):
        if not isinstance(value, uxr.Grid):
            raise TypeError("grid must be an instance of uxarray.Grid")
        self._grid = value
    
class IcosphereGrid(GridStrategy):    
    """
    Create a uniform grid configuration using an icosphere. This is the most accurate and efficient way to create a uniform grid, but is limited to a few resolutions.
    
    Parameters
    ----------
    gridlevel : float
        The subdivision level of the icosphere. The number of faces is 20 * 4**level. The default level is 8.
    radius: FloatLike
        The radius of the target body in meters.
        
    Returns
    -------
    IcosphereGrid
        An instance of the IcosphereGrid class initialized with the given pixel size. 
    """    
    
    def __init__(self, 
                 gridlevel: int = 8, 
                 radius: FloatLike = 1.0, 
                 **kwargs: Any):
        super().__init__(**kwargs)
        self.gridlevel = gridlevel
        self.radius = radius
        
        
    def generate_face_distribution(self) -> NDArray:
        """
        Creates the points that define the mesh centers.
           
        Returns
        -------
        (3,n) ndarray of np.float64
            Array of points on a unit sphere.
        """ 
        from trimesh.creation import icosphere
       
        print(f"Generating a mesh with icosphere level {self.gridlevel}.")  
        mesh = icosphere(self.gridlevel)
        points = mesh.vertices.T
        return points
   
    
    def generate_grid(self,
                      grid_file: os.PathLike,
                      grid_hash: str | None = None,
                      **kwargs: Any) -> Tuple[os.PathLike, os.PathLike]:        
        super().generate_grid(grid_file=grid_file, grid_hash=grid_hash, **kwargs)
        face_areas = self.grid.face_areas 
        face_sizes = np.sqrt(face_areas / (4 * np.pi))
        pix_mean = face_sizes.mean().item() * self.radius
        pix_std = face_sizes.std().item() * self.radius
        print(f"Effective pixel size: {pix_mean:.2f} +/- {pix_std:.2f} m")
        return    
    
class ArbitraryResolutionGrid(GridStrategy):
    """
    Create a uniform grid configuration with an arbitrary user-defined pixel size. This will not be as nice as the regular IcosphereGrid, but can be any resolution desired.
    
    Parameters
    ----------
    pix : float
        The approximate face size for the mesh in meters.
    radius: FloatLike
        The radius of the target body in meters.
        
    Returns
    -------
    ArbitraryResolutionGrid
        An instance of the ArbitraryResolutionGrid class initialized with the given pixel size. 
    """    
    
    def __init__(self, 
                 pix: FloatLike, 
                 radius: FloatLike, 
                 **kwargs: Any):
        super().__init__(**kwargs)
        self.pix = pix
        self.radius = radius
    

    def generate_face_distribution(self) -> NDArray:
        """
        Creates the points that define the mesh centers.
           
        Returns
        -------
        (3,n) ndarray of np.float64
            Array of points on a unit sphere.
        
        """                

        print(f"Generating a mesh with uniformly distributed faces of size ~{self.pix} m.")
        points = distribute_points(distance=self.pix/self.radius) 
        points[:,0] = np.array([0,0,1])
        points[:,-1] = np.array([0,0,-1])
        return points
   
    
    def generate_grid(self,
                      grid_file: os.PathLike,
                      grid_hash: str | None = None,
                      **kwargs: Any) -> Tuple[os.PathLike, os.PathLike]:        
        super().generate_grid(grid_file=grid_file, grid_hash=grid_hash, **kwargs)
        face_areas = self.grid.face_areas 
        face_sizes = np.sqrt(face_areas / (4 * np.pi))
        pix_mean = face_sizes.mean().item() * self.radius
        pix_std = face_sizes.std().item() * self.radius
        print(f"Effective pixel size: {pix_mean:.2f} +/- {pix_std:.2f} m")
        return


class HiResLocalGrid(GridStrategy):
    """
    Create a uniform grid configuration with the given pixel size.
    
    Parameters
    ----------
    pix : FloatLike
        The approximate face size inside the local region in meters.
    radius: FloatLike
        The radius of the target body in meters.
    local_radius : FloatLike
        The radius of the local region in meters.
    local_location : PairOfFloats
        The longitude and latitude of the location in degrees.
    superdomain_scale_factor : FloatLike
        A factor defining the ratio of cell size to the distance from the local boundary. This is set so that smallest craters 
        that are modeled outside the local region are those whose ejecta could just reach the boundary.
        
    Returns
    -------
    HiResLocalGrid
        An instance of the HiResLocalGrid clas initialized with the given point distribution
    """        
    def __init__(self, 
                 pix: FloatLike, 
                 radius: FloatLike, 
                 local_radius: FloatLike, 
                 local_location: PairOfFloats,
                 superdomain_scale_factor: FloatLike,
                 **kwargs: Any):
        super().__init__(**kwargs)
        self.pix = pix
        self.radius = radius        
        self.local_radius = local_radius
        self.local_location = local_location
        self.superdomain_scale_factor = superdomain_scale_factor
        
    
    def _generate_variable_size_array(self) -> Tuple[NDArray, NDArray, NDArray]:
        """
        Create an array of target pixel sizes for pairs of longitude and latitude values for a high-resolution local mesh around a 
        given location, with adaptive cell sizes outside the local region.
        
            
        Returns
        -------
        pix_array: ndarray
            m x n array of pixel sizes
        lon : ndarray
            longitude in degrees (length n and between -180 and 180)
        lat : ndarray
            latitude in degrees (length m and between -90 and 90)
        """
        
        def _pix_func(lon,lat):
            lon_rad = np.radians(lon)
            lat_rad = np.radians(lat)
            loc_lon_rad = 0.0 #np.radians(self.local_location[0])
            loc_lat_rad = 0.0 #np.radians(self.local_location[1])

            # Calculate distance from the location to the grid point
            distance = Surface.calculate_haversine_distance(loc_lon_rad, loc_lat_rad, lon_rad, lat_rad, self.radius)
            ans = np.where(distance <= self.local_radius, self.pix, (distance - self.local_radius) / self.superdomain_scale_factor + self.pix)
            return ans
        
        from scipy.interpolate import interp1d

        # Suppose we know pix(lat, lon)
        # Step 1: Construct a fine preliminary grid to estimate integrals
       
        Lat = np.linspace(-90., 90., 733)
        Lon = np.linspace(-180., 180., 733)
        LAT, LON = np.meshgrid(Lat, Lon, indexing='ij')

        pix_values = _pix_func(LON, LAT)   # Evaluate pix on this fine grid
        w = 1.0 / pix_values

        # Step 2: Integrate w over longitude to get W_lat(lat)
        W_lat_vals = np.trapezoid(w, x=Lon, axis=1)  # integrate along lon dimension
        W_lat_cumulative = np.cumsum(W_lat_vals)
        W_lat_cumulative -= W_lat_cumulative[0] # normalize from 0 to 1
        W_lat_cumulative /= W_lat_cumulative[-1]  

        # Create a function to invert W_lat(lat)
        f_lat = interp1d(W_lat_cumulative, Lat, bounds_error=False, fill_value='extrapolate')
        
        M = int(2*np.pi*self.radius/pix_values.max()) - 1
        while M > 0:
            badval=False

            # Step 3: For each lat interval, choose lon lines similarly
            N = M + 1  # number of lon lines
            lat_lines = f_lat(np.linspace(0, 1, N))
            lon_lines = np.zeros((M, N))

            for i in range(M):
                lat_low, lat_high = lat_lines[i], lat_lines[i+1]
                # Extract w in this lat band
                mask = (LAT >= lat_low) & (LAT <= lat_high)
                w_band = w[mask].reshape(-1, len(Lon))  
                # Integrate this band over lat
                w_band_vals = np.trapezoid(w_band, x=Lat[(Lat>=lat_low)&(Lat<=lat_high)], axis=0)
                W_lon_band_cumulative = np.cumsum(w_band_vals)
                if W_lon_band_cumulative[-1] > 0:
                    W_lon_band_cumulative -= W_lon_band_cumulative[0]
                    W_lon_band_cumulative /= W_lon_band_cumulative[-1]
                    f_lon = interp1d(W_lon_band_cumulative, Lon, bounds_error=False, fill_value='extrapolate')
                    lon_lines[i,:] = f_lon(np.linspace(0, 1, N))
                else:
                    badval=True
                    M = M // 2
                    break
                
            if not badval:
                break
            
        if badval:
            raise ValueError("Could not generate a grid with the given parameters. Please try again with different parameters.")
            
        LAT = np.zeros((M, N))
        LON = np.zeros((M, N))
        
        for i in range(M):
            LAT[i, :] = lat_lines[i]  # every point on this horizontal line has the same lat

        LON = lon_lines
            
        pix_array = _pix_func(LON, LAT) 

        return pix_array, LON, LAT

    def _rotate_point_cloud(self,points):
        """
        Rotate a point cloud so that the point at [-r,0,0] moves to (lon,lat)
        using the convention:
        - Longitude [-180, 180] degrees, increasing eastward
        - Latitude [-90, 90] degrees, increasing northward
        - (0,0) lon/lat corresponds to [r,0,0] in Cartesian space
        
        Parameters:
            points (np.ndarray): Nx3 array of (x,y,z) points.
            lon (float): Target longitude in degrees.
            lat (float): Target latitude in degrees.
            r (float): Radius of the sphere (default=1).
        
        Returns:
            np.ndarray: Rotated Nx3 point cloud.
        """

        from scipy.spatial.transform import Rotation as R
        # Convert target lon, lat to radians
        lon_rad, lat_rad = np.radians(self.local_location)
        
        # Compute target unit vector (correcting for lon,lat convention)
        target = np.array([
            self.radius * np.cos(lat_rad) * np.cos(lon_rad),  
            self.radius * np.cos(lat_rad) * np.sin(lon_rad),  
            self.radius * np.sin(lat_rad)                     
        ])

        # Original vector (the point we want to move)
        original = np.array([-self.radius, 0, 0])  # Starts at [-r, 0, 0]
        if np.isclose(lon_rad, 0.0) and np.isclose(lat_rad, 0.0):
            rotation = R.from_euler('z', 180, degrees=True)  # 180-degree rotation around z-axis
            return rotation.apply(points)

        # Compute the axis of rotation (cross product)
        axis = np.cross(original, target)
        axis_norm = np.linalg.norm(axis)
        
        # If the axis is zero (no rotation needed), return the original points
        if axis_norm < 1e-10:
            return points

        axis /= axis_norm  # Normalize the axis

        # Compute the rotation angle (dot product)
        angle = np.arccos(np.clip(np.dot(original / self.radius, target / self.radius), -1.0, 1.0))  # Normalize for dot product

        # Create rotation object
        rotvec = axis * angle  # Convert to rotation vector
        rotation = R.from_rotvec(rotvec)  # Create rotation from axis-angle

        return rotation.apply(points)

    def generate_face_distribution(self) -> NDArray:
        """
        Creates the points that define the mesh centers.
           
        Returns
        -------
        (3,n) ndarray of np.float64
            Array of points on a unit sphere.
        
        """                
        
        print(f"Generating a mesh with variable resolution faces")
        print(f"Center of local region: {self.local_location}")
        print(f"Size of local region: {self.local_radius:.2f} m")
        print(f"Hires region pixel size: {self.pix:.2f} m")
        print(f"Lores region pixel size: {self.pix * self.superdomain_scale_factor:.2f} m")
        pix_array, lon, lat = self._generate_variable_size_array()
        
        points = []
        n = lon.shape[0]
        m = lat.shape[1]
       
        for i in range(n-1):
            for j in range(m-1):
                lon_range = (lon[i,j], lon[i,j+1])
                lat_range = (lat[i,j], lat[i+1,j])
                p = distribute_points(distance=pix_array[i,j]/self.radius, lon_range=lon_range, lat_range=lat_range) 
                if p is not None:
                    points.append(p)
                
        points = np.concatenate(points, axis=1)
        points = self._rotate_point_cloud(points.T).T
        
        return points


    def generate_grid(self,
                      grid_file: os.PathLike,
                      grid_hash: str | None = None,
                      **kwargs: Any) -> Tuple[os.PathLike, os.PathLike]:        
        super().generate_grid(grid_file=grid_file, grid_hash=grid_hash, **kwargs)
        face_areas = self.grid.face_areas 
        face_sizes = np.sqrt(face_areas / (4 * np.pi))
        pix_min = face_sizes.min().item() * self.radius
        pix_max = face_sizes.max().item() * self.radius
        print(f"Generated {self.grid.n_face} faces")
        print(f"Effective pixel size range: {pix_min:.2f},{pix_max:.2f} m")
        return

    @property
    def local_radius(self):
        """
        The radius of the local region in meters.
        """
        return self._local_radius

    @local_radius.setter
    def local_radius(self, value: FloatLike):
        if not isinstance(value, FloatLike) or np.isnan(value) or np.isinf(value) or value <= 0:
            raise TypeError("local_radius must be a positive float")
        if value > 2 * np.pi * self.radius:
            raise ValueError("local_radius must be less than 2 * pi * radius of the target body")
        self._local_radius = value
        
    @property
    def local_location(self):
        """
        The longitude and latitude of the location in degrees.
        """
        return self._local_location
    
    @local_location.setter
    def local_location(self, value: PairOfFloats):
        if not isinstance(value, tuple) or len(value) != 2:
            raise TypeError("local_location must be a tuple of two floats")
        self._local_location = validate_and_convert_location(value)
        
    @property
    def superdomain_scale_factor(self):
        """
        A factor defining the ratio of cell size to the distance from the local boundary. This is set so that smallest craters that are 
        modeled outside the local region are those whose ejecta could just reach the boundary.
        """
        return self._superdomain_scale_factor
    
    @superdomain_scale_factor.setter
    def superdomain_scale_factor(self, value: FloatLike):
        if not isinstance(value, FloatLike) or np.isnan(value) or np.isinf(value) or value < 1.0:
            raise TypeError("superdomain_scale_factor must be a positive float greater than or equal to 1")
        self._superdomain_scale_factor = value
   
    
class Surface(UxDataset):
    """
    This class is used for handling surface-related data and operations in the cratermaker project. It provides methods for 
    setting elevation data, calculating distances and bearings, and other surface-related computations.
    
    The Surface class extends UxDataset for the cratermaker project.
    
    Parameters
    ----------
    *args
        Variable length argument list for additional parameters to pass to the ``uxarray.UxDataset`` class.
    target : Target, optional
        The target body or name of a known target body for the impact simulation. 
    data_dir : os.PathLike, optional
        The directory for data files.
    grid_file : os.PathLike, optional
        The file path to the grid file.
    compute_face_areas : bool, optional
        Flag to indicate whether to compute face areas. Default is False.    
    rng : Generator, optional
        A random number generator instance. If not provided, the default numpy RNG will be used.        
    **kwargs
        This is used to pass additional keyword arguments to pass to the ``uxarray.UxDataset`` class.
    """
    __slots__ = UxDataset.__slots__ + ('_name', '_description', '_data_dir', '_grid_file', '_smallest_length', '_area', '_target', '_rng')    

    def __init__(self, 
                 *args, 
                 target: Target | None = None, 
                 data_dir: os.PathLike | None = None,
                 grid_file: os.PathLike | None = None,
                 compute_face_areas: bool = False,
                 rng: Generator | None = None, 
                 **kwargs):

        # Call the super class constructor with the UxDataset
        super().__init__(*args, **kwargs)
       
        # Additional initialization for Surface
        self._name = "Surface"
        self._description = "Surface class for cratermaker"
        self._data_dir = data_dir
        self._grid_file = grid_file
        self._target = target
        self._area = None
        self._smallest_length = None
        self.rng = rng
       
        if compute_face_areas: 
            # Compute face area needed in the non-normalized units for future calculations
            self['face_areas'] = self.uxgrid.face_areas.assign_attrs(units='m^2') * self.target.radius**2
            self.smallest_length = np.sqrt(self['face_areas'].min().item()) * _SMALLFAC     
        
        return   

    @classmethod
    def initialize(cls, 
                   target: Target | None,
                   data_dir: os.PathLike | None = None,
                   grid_file: os.PathLike | None = None,
                   reset_surface: bool = True, 
                   grid_type: GridType = valid_grid_types[0], 
                   rng: Generator | None = None,
                   regrid: bool = False,
                   **kwargs):
        """
        Factory method to create a Surface instance from a grid file.

        Parameters
        ----------
        target : Target, optional
            Target object or name of known body for the simulation. Default is Target("Moon")
        data_dir : os.PathLike
            The directory for data files. Default is set to ``${PWD}/surface_data``.
        grid_file : os.PathLike
            The file path to the grid file. Default is set to be from the current working directory, to ``{data_dir}/grid.nc``.
        reset_surface : bool, optional
            Flag to indicate whether to reset the surface. Default is True.
        grid_type : ["icosphere", "arbitrary", "hires local"], optional
            The type of grid to be generated. Default is "icosphere".            
        rng : Generator, optional
            A random number generator instance. If not provided, the default numpy RNG will be used. 
        **kwargs : dict
            Additional keyword arguments for initializing the Surface instance based on the specific grid_type.

        Returns
        -------
        Surface
            An initialized Surface object.
                   
        Notes
        -----
        The grid configuration is determined by `grid_type` and its associated parameters. For detailed information on the 
        parameters specific to each grid type, refer to the documentation of the respective grid parameter classes (`IcosphereGrid`,`ArbitraryResolutionGrid`, 
        `HiResLocalGrid`, etc.).

        See Also
        --------
        cratermaker.core.surface.IcosphereGrid : Parameters for an icosphere grid configuration.
        cratermaker.core.surface.ArbitraryResolutionGrid : Parameters for a uniform arbitrary resolution grid configuration.
        cratermaker.core.surface.HiResLocalGrid : Parameters for a high-resolution local grid configuration.  
        """
        if not target:
            target = Target("Moon")
        elif isinstance(target, str):
            try:
                target = Target(target)
            except:
                raise ValueError(f"Invalid target name {target}")
        elif not isinstance(target, Target):
            raise TypeError("target must be an instance of Target or a valid name of a target body")
        
        if grid_type not in valid_grid_types:
            raise ValueError(f"Invalid grid_type {grid_type}. Valid options are {valid_grid_types}")
        
        gridlevel = kwargs.pop('gridlevel', 8)
         
        pix = kwargs.pop('pix', None) 
        if pix is not None:
            pix = np.float64(pix)
        else:    
            pix = np.sqrt(4 * np.pi * target.radius**2) * 1e-3  # Default mesh scale that is somewhat comparable to a 1000x1000 CTEM grid

        # Verify directory structure exists and create it if not
        if not data_dir:
            data_dir = os.path.join(os.getcwd(), _DATA_DIR)
            
        if not os.path.exists(data_dir):
            os.mkdir(data_dir)
            reset_surface = True
      
        if not grid_file: 
            grid_file = os.path.join(data_dir,_GRID_FILE_NAME)
            
        # Process the grid parameters from the arguments and build the strategy object 
        if grid_type == "icosphere":
            grid_strategy = IcosphereGrid(gridlevel=gridlevel, radius=target.radius)
        if grid_type == "arbitrary":
            grid_strategy = ArbitraryResolutionGrid(pix=pix, radius=target.radius)
        elif grid_type == "hires local":
            grid_strategy = HiResLocalGrid(pix=pix, radius=target.radius, **kwargs)       
   
        # Check if a grid file exists and matches the specified parameters based on a unique hash generated from these parameters. 
        if not regrid: 
            make_new_grid = grid_strategy.check_if_regrid(grid_file=grid_file, **kwargs)
        else:
            make_new_grid = True
        
        if make_new_grid:
            print("Creating a new grid")
            grid_strategy.create_grid(grid_file=grid_file, **kwargs)
        else:
            print("Using existing grid")
        
        # Get the names of all data files in the data directory that are not the grid file
        data_file_list = glob(os.path.join(data_dir, "*.nc"))
        if grid_file in data_file_list:
            data_file_list.remove(grid_file)
            
        # Generate a new surface if either it is explicitly requested via parameter or a data file doesn't yet exist 
        reset_surface = reset_surface or make_new_grid or not data_file_list
        
        # If reset_surface is True, delete all data files except the grid file 
        if reset_surface:
            for f in data_file_list:
                os.remove(f)
            data_file_list = []        
        
        # Initialize UxDataset with the loaded data
        try:
            if data_file_list:
                surf = uxr.open_mfdataset(grid_file, data_file_list, use_dual=False).isel(time=-1)
                surf.uxgrid = uxr.open_grid(grid_file, use_dual=False)
            else:
                surf = uxr.UxDataset()
                surf.uxgrid = uxr.open_grid(grid_file, use_dual=False)
        except:
            raise ValueError("Error loading grid and data files")
        
        surf = cls(surf,
                   uxgrid=surf.uxgrid,
                   source_datasets=surf.source_datasets,
                   target = target,
                   data_dir = data_dir,
                   grid_file = grid_file,
                   rng = rng,
                   compute_face_areas = True,
                  ) 
        
        if reset_surface:
            surf.generate_data(data=0.0,
                               name="ejecta_thickness",
                               long_name="ejecta thickness",
                               units= "m",
                               save_to_file=True
                              )     
            surf.generate_data(data=0.0,
                               name="ray_intensity",
                               long_name="ray intensity value",
                               units= "",
                               save_to_file=True
                              )                         
            surf.set_elevation(0.0,save_to_file=True)
        
        return surf        
        
    def __getitem__(self, key):
        """Override to make sure the result is an instance of ``cratermaker.Surface``"""

        value = super().__getitem__(key)

        if isinstance(value, uxr.UxDataset):
            value = Surface(value,
                            uxgrid=self.uxgrid,
                            source_datasets=self.source_datasets,
                            target=self.target,
                            grid_file=self.grid_file,
                            data_dir=self.data_dir,
                            rng=self.rng,
                            compute_face_areas=False,
                            )
        return value
    
    def _calculate_binary_op(self, *args, **kwargs):
        """Override to make the result a complete instance of ``cratermaker.Surface``."""
        ds = super()._calculate_binary_op(*args, **kwargs)

        if isinstance(ds, Surface):
            ds._name = self._name
            ds._description = self._description
            ds._data_dir = self._data_dir
            ds._grid_file = self._grid_file
            ds._target = self._target
            ds._rng = self._rng
            ds._smallest_length = self._smallest_length
        else:
            ds = Surface(ds,
                         uxgrid=self.uxgrid,
                         source_datasets=self.source_datasets,
                         target=self.target,
                         grid_file=self.grid_file,
                         data_dir=self.data_dir,
                         rng=self.rng,
                         compute_face_areas=False,
                         )
        return ds
    
    @classmethod
    def _construct_direct(cls, *args, **kwargs):
        """Override to make the result a ``cratermaker.Surface`` class."""

        return cls(uxr.UxDataset._construct_direct(*args, **kwargs))    

    def _copy(self, **kwargs):
        """Override to make the result a complete instance of ``cratermaker.Surface``."""
        copied = super()._copy(**kwargs)

        copied._name = self._name
        copied._description = self._description
        copied._data_dir = self._data_dir
        copied._grid_file = self._grid_file
        copied._smallest_length = self._smallest_length
        copied._area = self._area
        copied._target = self._target
        copied._rng = self._rng
        
        return copied    
  
    def _replace(self, *args, **kwargs):
        """Override to make the result a complete instance of ``cratermaker.Surface``."""
        ds = super()._replace(*args, **kwargs)

        if isinstance(ds, Surface):
            ds._name = self._name
            ds._description = self._description
            ds._data_dir = self._data_dir
            ds._grid_file = self._grid_file
            ds._target = self._target
            ds._rng = self._rng
            ds._smallest_length = self._smallest_length
            ds._area = self._area
        else:
            ds = Surface(ds,
                         uxgrid=self.uxgrid,
                         source_datasets=self.source_datasets,
                         target=self.target,
                         grid_file=self.grid_file,
                         data_dir=self.data_dir,
                         compute_face_areas=False,
                         rng=self.rng,
                         )
            ds._smallest_length = self._smallest_length
            ds._area = self._area
        return ds   

    @property
    def data_dir(self):
        """
        Directory for data files.
        """
        return self._data_dir

    @data_dir.setter
    def data_dir(self, value):
        self._data_dir = value
        if not os.path.exists(self._data_dir):
            os.makedirs(self._data_dir)

    @property
    def grid_file(self):
        """
        Path to the grid file.
        """
        return self._grid_file

    @grid_file.setter
    def grid_file(self, value):
        self._grid_file = value
        
    @property
    def smallest_length(self):
        """
        Smallest length value that is directly modeled on the grid. This is used to determine the maximum distance of ejecta to 
        consider, for instance
        """
        return self._smallest_length

    @smallest_length.setter
    def smallest_length(self, value):
        if not isinstance(value, float):
            raise TypeError("smallest_length must be a float")
        self._smallest_length = value
        
    @property
    def area(self):
        """
        Total surface area of the target body.
        """
        if self._area is None:
            self._area = self['face_areas'].sum().assign_attrs({'long_name': 'total surface area', 'units': 'm^2'})
        return self._area

    @property
    def target(self):
        """
        The target body for the impact simulation. Set during initialization.
        """
        return self._target

    @target.setter
    def target(self, value):
        if not isinstance(value, Target):
            raise TypeError("target must be an instance of Target")
        self._target = value    
        
    @property
    def rng(self):
        """
        A random number generator instance.
        
        Returns
        -------
        Generator
        """ 
        return self._rng
    
    @rng.setter
    def rng(self, value):
        if not isinstance(value, Generator) and value is not None:
            raise TypeError("The 'rng' argument must be a numpy.random.Generator instance or None")
        self._rng = value or np.random.default_rng()           
        
    def generate_data(self,
                      name: str,
                      long_name: str | None = None,
                      units: str | None = None,
                      data: FloatLike | NDArray | None = None,
                      isfacedata: bool = True,
                      save_to_file: bool = False,
                      interval_number: int = 0,
                      combine_data_files: bool = False,
                    ) -> None:
        """
        Generate either a node or face data variable and optionally save it to a file. If the data variable already exists, it will be overwritten.

        Parameters
        ----------
        name : str
            Name of the data variable. This will also be used as the data file name.
        long_name : str, optional
            Long name of the data variable that will be saved as an attribute.
        units : str, optional
            Units of the data variable that will be saved as an attribute.
        data : scalar or array-like
            Data file to be saved. If data is a scalar, then the data file will be filled with that value. If data is an array, then the data file will be filled with the array values. The data array must have the same size as the number of faces or nodes in the grid.
        isfacedata : bool, optional
            Flag to indicate whether the data is face data or node data. Default is True.
        save_to_file: bool, optional
            Specify whether the data should be saved to a file. Default is False.
        interval_number : int, optional, default 0
            The interval number to use when saving the data to the data file.
        combine_data_files : bool, optional
            If True, combine the current data with the existing data for previous intervals in the data file. Default is False.
            
        Returns
        -------
        None
        """    
        uxgrid = uxr.open_grid(self.grid_file,use_dual=False)
        if long_name is None and units is None:
            attrs = None
        else:
            attrs = {}
            if long_name:
                attrs["long_name"] = long_name
            if units:
                attrs["units"] = units
                
        if isfacedata: 
            dims = ["n_face"]
            size = uxgrid.n_face
        else:
            dims = ["n_node"]
            size = uxgrid.n_node
    
        if data is None:
            data = np.zeros(size,dtype=np.float64) 
        elif np.isscalar(data):
            data = np.full(size,data)
        else:
            if data.size != size:
                raise ValueError("data must have the same size as the number of faces or nodes in the grid") 
        uxda = UxDataArray(
            data=data,
            dims=dims,
            attrs=attrs,
            name=name,
            uxgrid=uxgrid,
        ) 
         
        self[name] = uxda
        
        if save_to_file:
            _save_data(uxda, self.data_dir, interval_number, combine_data_files)
        return 
        
    def set_elevation(self, 
                    new_elev: NDArray[np.float64] | List[FloatLike] | None = None,
                    save_to_file: bool = False, 
                    combine_data_files: bool = False,
                    interval_number: int = 0,
                    ) -> None:
        """
        Set elevation data for the target's surface mesh.

        Parameters
        ----------
        new_elev : array_like, optional
            New elevation data to be set. If None, the elevation is set to zero. 
        save_to_file : bool, default False
            If True, save the elevation data to the elevation file.
        combine_data_files : bool, default False
            If True, combine the current data with the existing data for previous intervals in the data file.
        interval_number : int, default 0
            The interval number to use when saving the elevation data to the elevation file.
        """
        if new_elev is None or np.isscalar(new_elev):
            gen_node = True
            gen_face = True 
        elif new_elev.size == self.uxgrid.n_node:
            gen_node = True
            gen_face = False
        elif new_elev.size == self.uxgrid.n_face:
            gen_node = False
            gen_face = True
        else:
            gen_node = False
            gen_face = False
            raise ValueError("new_elev must be None, a scalar, or an array with the same size as the number of nodes in the grid")
    
        if gen_node:
            self.generate_data(data=new_elev, 
                               name="node_elevation",
                               long_name="elevation of nodes",
                               units= "m",
                               isfacedata=False,
                               save_to_file=save_to_file,
                               combine_data_files=combine_data_files,
                               interval_number=interval_number
                              )   
        if gen_face:
            self.generate_data(data=new_elev, 
                               name="face_elevation",
                               long_name="elevation of faces",
                               units="m",
                               isfacedata=True,
                               save_to_file=save_to_file,
                               combine_data_files=combine_data_files,
                               interval_number=interval_number
                              )
        return 

    @staticmethod
    def calculate_haversine_distance(lon1: FloatLike, 
                    lat1: FloatLike, 
                    lon2: FloatLike, 
                    lat2: FloatLike,
                    radius: FloatLike = 1.0) -> np.float64:
        """
        Calculate the great circle distance between two points on a sphere.

        Parameters
        ----------
        lon1 : FloatLike
            Longitude of the first point in radians.
        lat1 : FloatLike
            Latitude of the first point in radians.
        lon2 : FloatLike
            Longitude of the second point in radians.
        lat2 : FloatLike
            Latitude of the second point in radians.
        radius : FloatLike
            Radius of the sphere in meters.

        Returns
        -------
        np.float64
            Great circle distance between the two points in meters.
        """
        # Calculate differences in coordinates
        dlon = lon2 - lon1
        dlat = lat2 - lat1

        # Haversine formula
        a = np.sin(dlat/2.0)**2 + np.cos(lat1) * np.cos(lat2) * np.sin(dlon/2.0)**2
        c = 2 * np.arcsin(np.sqrt(a))
        return radius * c
    
    def get_distance(self, 
                     location: Tuple[np.float64, np.float64]) -> UxDataArray:
        """
        Computes the distances between nodes and faces and a given location.

        Parameters
        ----------
        location : Tuple[np.float64, np.float64]
            Tuple containing the longitude and latitude of the location in degrees.

        Returns
        -------
        UxDataArray
            DataArray of distances for each node in meters.
        UxDataArray
            DataArray of distances for each face in meters.
        """
        lon1 = np.deg2rad(location[0])
        lat1 = np.deg2rad(location[1])
        node_lon2 = np.deg2rad(self.uxgrid.node_lon)
        node_lat2 = np.deg2rad(self.uxgrid.node_lat)        
        face_lon2 = np.deg2rad(self.uxgrid.face_lon)
        face_lat2 = np.deg2rad(self.uxgrid.face_lat)
        return self.calculate_haversine_distance(lon1,lat1,node_lon2,node_lat2,self.target.radius), self.calculate_haversine_distance(lon1,lat1,face_lon2,face_lat2,self.target.radius) 

    @staticmethod
    def calculate_initial_bearing(lon1: FloatLike, 
                                lat1: FloatLike, 
                                lon2: FloatLike, 
                                lat2: FloatLike) -> np.float64:
        """
        Calculate the initial bearing from one point to another on the surface of a sphere.

        Parameters
        ----------
        lon1 : FloatLike
            Longitude of the first point in radians.
        lat1 : FloatLike
            Latitude of the first point in radians.
        lon2 : FloatLike
            Longitude of the second point in radians.
        lat2 : FloatLike
            Latitude of the second point in radians.

        Returns
        -------
        np.float64
            Initial bearing from the first point to the second point in radians.
        """
        # Calculate differences in coordinates
        dlon = np.mod(lon2 - lon1 + np.pi, 2 * np.pi) - np.pi

        # Haversine formula calculations
        x = np.sin(dlon) * np.cos(lat2)
        y = np.cos(lat1) * np.sin(lat2) - np.sin(lat1) * np.cos(lat2) * np.cos(dlon)
        initial_bearing = np.arctan2(x, y)

        # Normalize bearing to 0 to 2*pi
        initial_bearing = (initial_bearing + 2 * np.pi) % (2 * np.pi)

        return initial_bearing
    
    def get_initial_bearing(self, location: Tuple[np.float64, np.float64]) -> UxDataArray:
        """
        Computes the initial bearing between nodes and faces and a given location.

        Parameters
        ----------
        location : Tuple[np.float64, np.float64]
            Tuple containing the longitude and latitude of the location in degrees.

        Returns
        -------
        DataArray
            DataArray of initial bearings for each node in radians.
        DataArray
            DataArray of initial bearings for each face in radians.
        """
        lon1 = np.deg2rad(location[0])
        lat1 = np.deg2rad(location[1])
        node_lon2 = np.deg2rad(self.uxgrid.node_lon)
        node_lat2 = np.deg2rad(self.uxgrid.node_lat)
        face_lon2 = np.deg2rad(self.uxgrid.face_lon)
        face_lat2 = np.deg2rad(self.uxgrid.face_lat)       
        return self.calculate_initial_bearing(lon1,lat1,node_lon2,node_lat2), self.calculate_initial_bearing(lon1,lat1,face_lon2,face_lat2)
    
    def find_nearest_index(self, location):
        """
        Find the index of the nearest node and face to a given point.

        This method calculates the Haversine distance from the given point to each face in the grid,
        and returns the index of the face with the minimum distance.

        Parameters
        ----------
        location : tuple
            A tuple containing two elements: (longitude, latitude) in degrees.

        Returns
        -------
        int
            The index of the nearest node in the grid to the given point.
        int
            The index of the nearest face in the grid to the given point.

        Notes
        -----
        The method uses the ball tree query method that is included in the UxArray.Grid class.
        """          
        
        coords = np.asarray(location)

        with warnings.catch_warnings():
            warnings.simplefilter("ignore", Warning)
            node_tree = self.uxgrid.get_ball_tree("nodes", distance_metric="haversine", coordinate_system="spherical")
            node_ind = node_tree.query(coords=coords, k=1, return_distance=False)
        
            face_tree = self.uxgrid.get_ball_tree("face centers",  distance_metric="haversine", coordinate_system="spherical")
            face_ind = face_tree.query(coords=coords, k=1, return_distance=False)
        return node_ind.item(), face_ind.item()

    def get_reference_surface(self,
                            location: Tuple[FloatLike, FloatLike], 
                            region_radius: np.float64) -> NDArray[np.float64]:
        """
        Calculate the orientation of a hemispherical cap that represents the average surface within a given region.

        Parameters
        ----------
        location : Tuple[float, float]
            Tuple containing the longitude and latitude of the reference location in degrees.
        region_radius : float
            The radius of the region to compute the average over in meters.

        Returns
        -------
        center_vector : ndarray
            The vector pointing to the center of the cap from the sphere's center.
        """
        def sphere_function(coords, x_c, y_c, z_c, r):
            """
            Compute the sphere function.

            Parameters
            ----------
            coords : ndarray
                Array of x, y, and z coordinates.
            x_c, y_c, z_c : float
                Center coordinates of the sphere.
            r : float
                Radius of the sphere.

            Returns
            -------
            ndarray
                The values of the sphere function at the given coordinates.
            """
            x, y, z = coords.T
            return (x - x_c)**2 + (y - y_c)**2 + (z - z_c)**2 - r**2
   
        # Find cells within the crater radius
        if 'face_crater_distance' and 'node_crater_distance' in self:
            faces_within_radius = self['face_crater_distance'] <= region_radius
            nodes_within_radius = self['node_crater_distance'] <= region_radius
        else:
            nodes_within_radius, faces_within_radius = self.get_distance(location)
            faces_within_radius = faces_within_radius <= region_radius
            nodes_within_radius = nodes_within_radius <= region_radius
        
        if np.sum(faces_within_radius) < 5 or not nodes_within_radius.any():
            self['reference_face_elevation'] = self['face_elevation']
            self['reference_node_elevation'] = self['node_elevation']
            return
       
        inc_face = self.n_face
        face_vars = ['face_x', 'face_y', 'face_z'] 
        face_grid = self.n_face.uxgrid._ds[face_vars].sel(n_face=inc_face)
        region_faces = face_grid.where(faces_within_radius, drop=False)
        region_elevation = self['face_elevation'].where(faces_within_radius, drop=False) / self.target.radius
        region_surf = self.elevation_to_cartesian(region_faces, region_elevation) 

        x, y, z = region_surf['face_x'], region_surf['face_y'], region_surf['face_z']
        region_vectors = np.vstack((x, y, z)).T

        # Initial guess for the sphere center and radius
        guess_radius = 1.0 + region_elevation.mean().values.item() 
        initial_guess = [0, 0, 0, guess_radius]  

        # Perform the curve fitting
        with warnings.catch_warnings():
            warnings.simplefilter("ignore", OptimizeWarning)
            try:
                bounds =([-1.0, -1.0, -1.0, 0.5],[1.0, 1.0, 1.0, 2.0]) 
                popt, _  = curve_fit(sphere_function, region_vectors[faces_within_radius], np.zeros_like(x[faces_within_radius]), p0=initial_guess, bounds=bounds)
            except:
                popt = initial_guess

        # Extract the fitted sphere center and radius
        reference_sphere_center = popt[:3]
        reference_sphere_radius = popt[3]
        
        def find_reference_elevations(coords):
            # Find the point along the original vector that intersects the sphere 
            f_vec = coords / self.target.radius      
            A = f_vec[:,0]**2 + f_vec[:,1]**2 + f_vec[:,2]**2
            B = -2 * (f_vec[:,0] * reference_sphere_center[0] + f_vec[:,1] * reference_sphere_center[1] + f_vec[:,2] * reference_sphere_center[2])
            C = np.dot(reference_sphere_center, reference_sphere_center) - reference_sphere_radius**2
            sqrt_term = B**2 - 4 * A * C
            valid = ~np.isnan(A) & (sqrt_term >= 0.0) 

            # Initialize t with default value
            t = np.full_like(A, 1.0)

            # Calculate square root only for valid terms
            sqrt_valid_term = np.sqrt(np.where(valid, sqrt_term, 0))

            # Apply the formula only where valid
            t = np.where(valid, (-B + sqrt_valid_term) / (2 * A), t)
            if np.any(t[valid] < 0):
                t = np.where(valid & (t < 0), (-B - sqrt_valid_term) / (2 * A), t)
                
            elevations = self.target.radius * (t * np.linalg.norm(f_vec, axis=1)  - 1)
            return elevations
        
        # Calculate the distances between the original face points and the intersection points
        face_vectors  = np.vstack((region_faces.face_x, region_faces.face_y, region_faces.face_z)).T
        self['reference_face_elevation'] = xr.where(faces_within_radius, find_reference_elevations(face_vectors), self['face_elevation'])
        
        # Now do the same thing to compute the nodal values 
        inc_node = self.n_node
        node_vars = ['node_x', 'node_y', 'node_z'] 
        node_grid = self.n_node.uxgrid._ds[node_vars].sel(n_node=inc_node)
        region_nodes = node_grid.where(nodes_within_radius, drop=False) 
        node_vectors  = np.vstack((region_nodes.node_x, region_nodes.node_y, region_nodes.node_z)).T
        
        self['reference_node_elevation'] = xr.where(nodes_within_radius, find_reference_elevations(node_vectors), self['node_elevation'])
        
        return

    @staticmethod
    def elevation_to_cartesian(position: Dataset, 
                            elevation: DataArray
                            ) -> Dataset:
        
        vars = list(position.data_vars)
        if len(vars) != 3:
            raise ValueError("Dataset must contain exactly three coordinate variables")
        
        dim_var = list(position.dims)[0]

        rvec = np.column_stack((position[vars[0]], position[vars[1]], position[vars[2]]))
        runit = rvec / np.linalg.norm(rvec, axis=1, keepdims=True)
        
        ds_new = Dataset(
                        {
                        vars[0]: ((dim_var,), rvec[:,0] + elevation.values * runit[:,0]),
                        vars[1]: ((dim_var,), rvec[:,1] + elevation.values * runit[:,1]),
                        vars[2]: ((dim_var,), rvec[:,2] + elevation.values * runit[:,2]),
                        }
                        )
        return ds_new    
   
    def extract_region(self,
                       location: Tuple[FloatLike, FloatLike],
                       region_radius: FloatLike): 
        
        """
        Extract a regional grid based on a given location and radius.
        
        Parameters
        ----------
        location : Tuple[float, float]
            Tuple containing the longitude and latitude of the location in degrees.
        region_radius : float
            The radius of the region to extract in meters.
            
        Returns
        -------
        Surface
            A new Surface object containing the regional grid. 
            
        Notes
        -----
        Though not well documented, the mapping from the region grid back to the original grid can be done using the variables
        "subgrid_face_indices" and "subgrid_node_indices" in the region grid. For example, the following will extract a region
        surface, modify the face elevation, then insert the modified values back onto the original surface:
        
        .. code-block:: python    
              
            import xarray as xr
                
            region_surf = surf.extract_region(location, region_radius)
            region_surf['face_elevation'] = xr.full_like(region_surf['face_elevation'], 1.0)
            surf['face_elevation'].loc[{'n_face': region_surf.uxgrid._ds["subgrid_face_indices"]}] = region_surf['face_elevation']
            
        """ 
        
        region_angle = np.rad2deg(region_radius / self.target.radius)
        try:
            region_grid = self.uxgrid.subset.bounding_circle(center_coord=location, r=region_angle,element="face centers")
        except ValueError:
            return None
        
        region_surf = self.isel(n_face=region_grid._ds["subgrid_face_indices"], n_node=region_grid._ds["subgrid_node_indices"])
        region_surf.uxgrid = region_grid
      
        return region_surf
    
    def get_random_location_on_face(self, 
                                    face_index: int, 
                                    size: int = 1,
                                    **kwargs
                                    ) -> Union[np.float64, Tuple[np.float64, np.float64], ArrayLike]:
        """
        Generate a random coordinate within a given face of an unstructured mesh.

        Parameters
        ----------
        grid : uxarray.Grid
            The grid object containing the mesh information.
        face_index : int
            The index of the face within the grid to obtain the random sample.
        size : int or tuple of ints, optional
            The number of samples to generate. If size is None (the default), a single tuple is returned. If size is greater than 1, 
            then a structured array with fields 'lon' and 'lat' is returned.
                
        Returns
        -------
        (lon,lat) or ndarray[(lon,lat)] of given size
            A pair or array of pairs of longitude and latitude values in degrees.

        Notes
        -----
        This method is a wrapper for :func:`cratermaker.utils.montecarlo.get_random_location_on_face`. 
        """
        
        return get_random_location_on_face(self.uxgrid, face_index, size,**kwargs)


def _save_data(ds: xr.Dataset | xr.DataArray,
               out_dir: os.PathLike,
               interval_number: int = 0,
               combine_data_files: bool = False
               ) -> None:
    """
    Save the data to the specified directory. If `combine_data_files` is True, then all data variables are saved to a single NetCDF
    file. If False, then only the data variables for the current interval are saved to a NetCDF file with the interval number
    appended. 
    
    Parameters
    ----------
    ds : xr.Dataset or xr.DataArray
        The data to be saved.
    out_dir : PathLike
        Directory to save the data.
    interval_number : int, Default is 0.
        Interval number to append to the data file name. Default is 0.
    combine_data_files : bool, Default is False.
        If True, combine all data variables into a single NetCDF file, otherwise each variable will be saved to its own NetCDF file. 
        
    Notes
    -----
    This function first saves to a temporary file and then moves that file to the final destination. This is done to avoid file 
    locking issues with NetCDF files.
    """
    if isinstance(ds, xr.DataArray):
        ds = ds.to_dataset()
        
    if "time" not in ds.dims:
        ds = ds.expand_dims(["time"])
    if "time" not in ds.coords:
        ds = ds.assign_coords({"time":[interval_number]})      
        
        
    with tempfile.TemporaryDirectory() as temp_dir:
        if combine_data_files:
            filename = _COMBINED_DATA_FILE_NAME
        else:
            filename = _COMBINED_DATA_FILE_NAME.replace(".nc", f"{interval_number:06d}.nc")
            
        data_file = os.path.join(out_dir, filename)
        if os.path.exists(data_file):
            ds_file = xr.open_mfdataset(data_file)
            ds_file = ds.merge(ds_file, compat="override")
        else:
            ds_file = ds    
            
        temp_file = os.path.join(temp_dir, filename)
        
        comp = dict(zlib=True, complevel=9)
        encoding = {var: comp for var in ds_file.data_vars}
        ds_file.to_netcdf(temp_file, encoding=encoding)
        ds_file.close()     
        shutil.move(temp_file, data_file)

    return


def distribute_points(distance: FloatLike,
                      radius: FloatLike=1.0, 
                      lon_range: PairOfFloats=(-180,180), 
                      lat_range: PairOfFloats=(-90,90)):
    """
    Distributes points on a sphere using Deserno's algorithm [1]_.
        
    Parameters
    ----------
    distance : float
        Approximate distance between points, used to determine the number of points, where n = 1/distance**2 when distributed over the whole sphere.
    radius : float, optional
        Radius of the sphere. Default is 1.0
    lon_range : tuple, optional
        Range of longitudes in degrees. Default is (-180,180).
    lat_range : tuple, optional
        Range of latitudes in degrees. Default is (-90,90).
        
    Returns
    -------
    (3,n) ndarray of np.float64
        Array of cartesian points on the sphere.
        
    
    References
    ----------
    .. [1] Deserno, Markus., 2004. How to generate equidistributed points on the surface of a sphere. https://www.cmu.edu/biolphys/deserno/pdf/sphere_equi.pdf
    
    """
    def _sph2cart(theta, phi, r):
        """ 
        Converts spherical coordinates to Cartesian coordinates.
        
        Parameters
        ----------
        theta : float
            Inclination angle in radians.
        phi : float
            Azimuthal angle in radians.
        r : float
            Radius.
        """
        x = r * np.sin(theta) * np.cos(phi)
        y = r * np.sin(theta) * np.sin(phi)
        z = r * np.cos(theta)
        return x, y, z
 
    
    phi_range = np.deg2rad(lon_range) + np.pi 
    theta_range = np.deg2rad(lat_range) + np.pi/2 
    points = []
       
    n = int(1/distance**2)
    if n < 1: 
        return 
        
    a = 4 * np.pi / n
    d = np.sqrt(a)
    Mtheta = int(np.round(np.pi / d))
    dtheta = np.pi / Mtheta
    dphi = a / dtheta
    
    thetavals = np.pi * (np.arange(Mtheta) + 0.5) / Mtheta
    thetavals = thetavals[(thetavals >= theta_range[0]) & (thetavals < theta_range[1])]
    
    for theta in thetavals:
        Mphi = int(np.round(2 * np.pi * np.sin(theta) / dphi))
        phivals = 2 * np.pi * np.arange(Mphi) / Mphi
        phivals = phivals[(phivals >= phi_range[0]) & (phivals < phi_range[1])]
        for phi in phivals:
            points.append(_sph2cart(theta, phi, radius))
    if len(points) == 0:
        return
    
    points = np.array(points,dtype=np.float64)
    points = points.T

    return points    


def save(surf: Surface, 
         out_dir: os.PathLike | None = None,
         combine_data_files: bool = False,
         interval_number: int = 0,
         time_variables: dict | None = None,
         *args, **kwargs, 
         ) -> None:
    """
    Save the surface data to the specified directory. Each data variable is saved to a separate NetCDF file. If 'time_variables' is specified, then a one or more variables will be added to the dataset along the time dimension. If 'interval_number' is included as a key in `time_variables`, then this will be appended to the data file name.

    Parameters
    ----------
    surface : Surface
        The surface object to be saved. 
    out_dir : str, optional
        Directory to save the surface data. If None, the data is saved to the current working directory.
    combine_data_files : bool, optional
        If True, combine all data variables into a single NetCDF file, otherwise each variable will be saved to its own NetCDF file. Default is False.
    interval_number : int, optional
        Interval number to append to the data file name. Default is 0.
    time_variables : dict, optional
        Dictionary containing one or more variable name and value pairs. These will be added to the dataset along the time dimension. Default is None.
    """
    do_not_save = ["face_areas"]
    if out_dir is None:
        out_dir = surf.data_dir
        
    if not os.path.exists(out_dir):
        os.makedirs(out_dir)         
      
    if time_variables is None:
        time_variables = {"elapsed_time":np.float64(interval_number)}  
    else:
        if not isinstance(time_variables, dict):
            raise TypeError("time_variables must be a dictionary")
        
    # Variables that we do not want to save as they are computed at runtime     
        
    surf.close()
    
    ds = surf.expand_dims(dim="time").assign_coords({"time":[interval_number]})
    for k, v in time_variables.items():
        ds[k] = xr.DataArray(data=[v], name=k, dims=["time"], coords={"time":[interval_number]})
                
    drop_vars = [k for k in ds.data_vars if k in do_not_save]
    if len(drop_vars) > 0:
        ds = ds.drop_vars(drop_vars)
        
    _save_data(ds, out_dir, interval_number, combine_data_files)

    return







import unittest
import cratermaker 
import tempfile
import os
import warnings

class TestRealistic(unittest.TestCase):

    def setUp(self):
        # Initialize a target and surface for testing
        self.temp_dir = tempfile.TemporaryDirectory()
        self.target = cratermaker.Target(name="Moon") 
        self.gridlevel = 4
        os.chdir(self.temp_dir.name) 
        
    def tearDown(self):
        # Clean up temporary directory
        self.temp_dir.cleanup() 
        return           

    def test_realistic(self):
        sim = cratermaker.Simulation(gridlevel=self.gridlevel)
        sim.apply_noise(model="ridged")

if __name__ == '__main__':
    unittest.main()



import numpy as np
from numpy.random import Generator
from numpy.typing import ArrayLike
from typing import Union, Optional, Tuple
from numpy.typing import NDArray
from scipy.stats import truncnorm
from scipy.stats import maxwell
from uxarray import Grid
import uxarray as uxr


def get_random_location(size: int=1, 
                        rng: Generator | None=None
                        ) -> Union[np.float64, Tuple[np.float64, np.float64], ArrayLike]:
    """
    Computes random longitude and latitude values.
    
    Generates a set of latitude and longitude values that are uniformly distributed on the surface of a sphere.
    
    Parameters
    ----------
    size : int or tuple of ints, optional
        The number of samples to generate. If size is None (the default), a single tuple is returned. If size is greater than 1, 
        then a structured array with fields 'lon' and 'lat' is returned.
    rng : numpy.random.Generator, optional
        An instance of a random number generator compatible with numpy's random generators. If not provided, `default_rng` is used 
        to create a new instance.
    
    Returns
    -------
    (lon,lat) or ndarray[(lon,lat)] of given size
        A pair or array of pairs of longitude and latitude values in degrees.
    """

    if rng and not isinstance(rng, Generator):
        raise TypeError("The 'rng' argument must be a numpy.random.Generator instance or None")
    if rng is None:
        rng = np.random.default_rng() 

    u = rng.uniform(size=size)
    v = rng.uniform(size=size)
    
    # Compute the angles theta and phi
    theta = 2 * np.pi * u
    phi = np.arccos(2 * v - 1)
    
    # Convert to lon/lat
    lon = np.rad2deg(theta - np.pi) # Use the convention that longitude is in the range [-180, 180]
    lat = np.rad2deg(phi - np.pi / 2.0)
    
    if size == 1: 
        return np.float64(lon.item()),np.float64(lat.item())
    else:
        # Reshape lat and lon to the original size if necessary
        lon = lon.reshape(size)
        lat = lat.reshape(size)
  
        # Combine lat and lon into a structured array
        lonlat_arr = np.empty(size, dtype=[('lon', 'float'), ('lat', 'float')])
        lonlat_arr['lon'] = lon
        lonlat_arr['lat'] = lat
    
    return lonlat_arr


def get_random_location_on_face(grid: Grid, 
                                face_index: int, 
                                size: int=1, 
                                rng: Generator | None=None
                                ) -> Union[np.float64, Tuple[np.float64, np.float64], ArrayLike]:
    """
    Generate a random coordinate within a given face of an unstructured mesh.

    Parameters
    ----------
    grid : uxarray.Grid
        The grid object containing the mesh information.
    face_index : int
        The index of the face within the grid to obtain the random sample.
    size : int or tuple of ints, optional
        The number of samples to generate. If size is None (the default), a single tuple is returned. If size is greater than 1, 
        then a structured array with fields 'lon' and 'lat' is returned.
    rng : numpy.random.Generator, optional
        An instance of a random number generator compatible with numpy's random generators. If not provided, `default_rng` is used
        
    Returns
    -------
    (lon,lat) or ndarray[(lon,lat)] of given size
        A pair or array of pairs of longitude and latitude values in degrees.
    """

    if rng and not isinstance(rng, Generator):
        raise TypeError("The 'rng' argument must be a numpy.random.Generator instance or None")
    if rng is None:
        rng = np.random.default_rng() 
        
    # Extract node indices for the given face
    node_indices = grid.face_node_connectivity[face_index, :]
    node_indices = node_indices[node_indices >= 0]
    
    # Initialize the array for storing locations
    locations = np.empty(size, dtype=[('lon', 'float64'), ('lat', 'float64')])

    for i in range(size):
        # Retrieve the x, y, z values for these nodes
        x = grid.node_x[node_indices]
        y = grid.node_y[node_indices]
        z = grid.node_z[node_indices]
        
        # Generate two random indices to define a triangle with the reference point
        idx1, idx2 = rng.choice(range(1, len(node_indices)), 2, replace=False)
        p0, p1, p2 = np.array([x[0], y[0], z[0]]), np.array([x[idx1], y[idx1], z[idx1]]), np.array([x[idx2], y[idx2], z[idx2]])
        
        # Generate random barycentric coordinates for interpolation within the triangle
        r1, r2 = rng.random(), rng.random()
        if r1 + r2 > 1:
            r1, r2 = 1 - r1, 1 - r2
        
        # Interpolate in Cartesian space
        p_random = r1 * p1 + r2 * p2 + (1 - r1 - r2) * p0
        
        # Convert the random Cartesian point back to lon/lat
        lon_lat = uxr.grid.coordinates._xyz_to_lonlat_deg(p_random[0],p_random[1],p_random[2])
        
        # Store the generated lon/lat values in the structured array
        locations['lon'][i] = lon_lat[0]
        locations['lat'][i] = lon_lat[1]

    # Return the appropriate format based on 'size'
    if size == 1:
        return locations['lon'][0], locations['lat'][0]
    else:
        return locations


def get_random_impact_angle(size: int | Tuple[int, ...]=1, 
                            rng: Generator | None=None
                            ) -> Union[np.float64,NDArray[np.float64]]:
    """
    Sample impact angles from a distribution centered on 45deg.
    
    For the theory, see Shoemaker (1962) "Interpretation of lunar craters."
    
    Parameters 
    ----------
    size : int or tuple of ints, optional
        The number of samples to generate. If the shape is (m, n, k), then m * n * k samples are drawn. If size is None (the default), a single value is returned if `diameters` is a scalar, otherwise an array of samples is returned with the same size as `diameters`.
    rng : numpy.random.Generator, optional
        An instance of a random number generator compatible with numpy's random generators. If not provided, `default_rng` is used to create a new instance.     
        
    Returns
    ----------
    np.float64 or ndarray of np.float64 
        A scalar or array of impact angles (in degrees).
    """    
    
    if rng and not isinstance(rng, Generator):
        raise TypeError("The 'rng' argument must be a numpy.random.Generator instance or None")
    if rng is None:
        rng = np.random.default_rng() 

    u = np.sqrt(rng.uniform(size=size))
    impact_angle = np.arcsin(u)
    if size == 1:
        return np.rad2deg(impact_angle[0])
    else:
        return np.rad2deg(impact_angle)


def get_random_size(diameters: NDArray[np.float64], 
                    cdf: NDArray[np.float64], 
                    size: int | Tuple[int, ...] | None = None, 
                    mu: int | Tuple[int, ...] | None = None,
                    rng: Generator | None=None
                    ) -> Union[np.float64,NDArray[np.float64]]:
    """
    Sample diameters from a cumulative size-frequency distribution (SFD).
    
    Given an array of diameters and optionally a cumulative distribution function (CDF), this function generates new diameter values that follow the specified distribution. The SFD is treated as a continuous function that interpolates between the provided diameters, which are assumed to represent a power-law distribution.
    
    Parameters
    ----------
    diameters : array_like
        An array of diameters from which the SFD is constructed. Must be 1-dimensional.
    cdf : array_like
        The cumulative distribution function corresponding to `diameters`. Must be the same size as `diameters` and must be monotonically increasing with decreasing diameter value.
    size : int or tuple of ints, optional
        The number of samples to generate. If the shape is (m, n, k), then m * n * k samples are drawn. If size is None and mu is None then a single value is returned. Note: mu and size are mutually exclusive. 
    mu : int or tuple of ints, optional
        The expected number of samples to generate using a Poisson random number genertor. If the shape is (m, n, k), then m * n * k samples are drawn. Note: mu and size are mutually exclusive. 
    rng : numpy.random.Generator, optional
        An instance of a random number generator compatible with numpy's random generators. If not provided, `default_rng` is used to create a new instance.
    
    Returns
    -------
    np.float64 or ndarray of np.float 64 
        A scalar or array of sampled diameter values from the SFD. 
    
    Notes
    -----
    The SFD is assumed to be a continuous distribution that follows a power-law between the provided discrete diameter values. Linear interpolation in log-space is used to sample new values between the known diameters.
    A small amount of random noise (of the order 1e-3 the diameter value) is added to the final diameter value to ensure that diameter values are unlikely to be identical, even when an input CDF could lead to identical diameter values.
    
    Examples
    --------
    >>> diameters = np.array([100.,  56.,  32.,  18.,  10.])
    >>> ncumul = np.array([1.  , 0.51, 0.21, 0.06, 0.01])
    >>> sample_from_sfd(diameters, cdf=ncumul, size=4)
    array([14.80803668, 44.95292261, 29.80797715, 23.11082091])
    
    See Also
    --------
    numpy.random.Generator : The numpy random generator class used for random sampling.
    
    """

    # Check if rng has 'uniform' method which is a characteristic of numpy's random generator objects and use that to generate our values
    if rng and not isinstance(rng, Generator):
        raise TypeError("The 'rng' argument must be a numpy.random.Generator instance or None")
    if rng is None:
        rng = np.random.default_rng() 
        
    # Check that the shapes and sizes of diameters and cdf are compatible
    if np.isscalar(diameters) or np.isscalar(cdf):
        raise ValueError("The 'diameters' and 'cdf' arguments must be arrays")
    if diameters.ndim != 1:
        raise ValueError("The 'diameters' argument must be a 1-dimensional array")
    if diameters.shape != cdf.shape:
        raise ValueError("The 'diameters' and 'cdf' arguments must have the same shape")
    if diameters.size != cdf.size:
        raise ValueError("The 'diameters' and 'cdf' arguments must have the same size")
    if diameters.size < 2:
        raise ValueError("The 'diameters' and 'cdf' arguments must have at least two elements")
    if np.any(diameters <= 0.0):
        raise ValueError("All values in the 'diameters' argument must be positive")
    if np.any(cdf <= 0.0):
        raise ValueError("All values in the 'cdf' argument must be positive")
    if size is None and mu is None:
        size = 1
    elif size is not None and mu is not None:
        raise ValueError("The 'size' and 'mu' arguments are mutually exclusive")
    elif size is None and mu is not None:
        size = rng.poisson(mu) 

    # Sort the diameters in descending order and get the cumulative distribution if it was not supplied
    sorted_indices = np.argsort(diameters)[::-1]
    sorted_diameters = diameters[sorted_indices]
    sorted_cdf = cdf[sorted_indices]
    
    # Check to make sure that the CDF is correctly specified so that as diameter is decreasing it is monotonically increasing
    is_monotonic_increasing = np.all(np.diff(sorted_cdf) >= 0)
    if not is_monotonic_increasing:
        raise ValueError("The CDF must be monotonically increasing with decreasing diameter.")
    
    # Normalize the cdf and put it in logspace
    sorted_cdf /= sorted_cdf[-1]
    log_sorted_cdf = np.log(sorted_cdf)

    # Generate uniform random numbers for the entire sample size
    u = rng.uniform(low=sorted_cdf[0], high=sorted_cdf[-1], size=size)
    u = np.log(u)

    # Handle the situation where u is a scalar (size is None and diameters is a scalar)
    if np.isscalar(u):
        u = np.array([u])
        
    # Flatten u to work with it as a 1D array
    original_shape = u.shape
    u = u.flatten()     

    # Find the indices where the random numbers would be inserted to maintain order
    # Use the right side of the interval to avoid edge effects for when u == sorted_cdf[0]
    # Because rng.uniform returns values in the half-open interval [sorted_cdf[0], sorted_cdf[-1]), u will never be exactly equal to sorted_cdf[-1]
    indices = np.searchsorted(log_sorted_cdf, u, side="right")

    # Initialize the new_diameters array
    new_diameters = np.empty(u.shape)

    log_diam_low = np.log(sorted_diameters[indices-1])
    log_diam_high = np.log(sorted_diameters[indices])
    log_nval_low = log_sorted_cdf[indices-1]
    log_nval_high = log_sorted_cdf[indices]
                             
    fractions = (u - log_nval_low) / (log_nval_high - log_nval_low)
    log_diam_interp = log_diam_low + fractions * (log_diam_high - log_diam_low)
    new_diameters = np.exp(log_diam_interp)
    
    # Reshape new_diameters to the original shape of u
    new_diameters = new_diameters.reshape(original_shape)

    # Add a small random noise to the diameters
    noise = 1e-8 * rng.uniform(size=new_diameters.shape)
    new_diameters *= (1 + noise)
    if size == 1:
        return new_diameters[0]
    else:
        return new_diameters


def get_random_velocity(vmean: np.float64, 
                        size: int | Tuple[int, ...]=1, 
                        rng: Generator | None=None
                        ) -> Union[np.float64,NDArray[np.float64]]:
    """
    Sample impact velocities from a Maxwell-Boltzmann distribution given a mean velocity.
    
    Parameters 
    ----------
    vmean : np.float64
        The mean velocity of the distribution.
    size : int or tuple of ints, optional
        The number of samples to generate. If the shape is (m, n, k), then m * n * k samples are drawn. If size is None (the default), a single value is returned if `diameters` is a scalar, otherwise an array of samples is returned with the same size as `diameters`.
    rng : numpy.random.Generator, optional
        An instance of a random number generator compatible with numpy's random generators. If not provided, `default_rng` is used to create a new instance.    
       
    Returns
    ----------
    ndarray 
        An array of impact velocities (in m/s).
    """
    
    # Check if rng has 'uniform' method which is a characteristic of numpy's random generator objects and use that to generate our values
    if rng and hasattr(rng, 'normal'):
        pass
    elif rng is None:  # Just use the basic normal random number generator
        rng = np.random.default_rng()
    else:
        raise TypeError("The 'rng' argument must be a compatible with numpy random generator or None")
    
    sigma = vmean / np.sqrt(8/np.pi)
    
    vx = rng.normal(0, sigma, size=size)
    vy = rng.normal(0, sigma, size=size)
    vz = rng.normal(0, sigma, size=size)
    velocities = np.sqrt(vx**2 + vy**2 + vz**2)
 
    if size == 1:
        return velocities[0]
    else:
        return velocities


def bounded_norm(loc: np.float64,scale: np.float64,size: Optional[Union[int, Tuple[int, ...]]]=1):
    """
    Sample from a truncated normal distribution that is bounded by 1-sigma stdev
    
    Parameters 
    ----------
    loc : float
       mean of the distribution
    scale : float
       standard deviation and bounds of the distribution
       
    Returns
    ----------
    float
       Truncated norm bounded by loc-scale, loc+scale
    """    
    
    lower_bound = loc - scale
    upper_bound = loc + scale
    truncated_normal = truncnorm(
          (lower_bound - loc) / scale,
            (upper_bound - loc) / scale,
            loc=loc, scale=scale
        )
    
    if size == 1:
        return truncated_normal.rvs(1)[0]
    else:
        return truncated_normal.rvs(size)
    
           
if __name__ == '__main__':
    import matplotlib.pyplot as plt

    def plot_random_location():
        # Sample data generation
        size = 10000
        points = get_random_location(size=size)
        
        lons = points['lon']
        lats = points['lat']

        # Number of bins
        bins = 50 

        # Longitude plot
        observed_counts_lon, bins_lon_deg = np.histogram(lons, bins=bins, range=(-180, 180.0))
        expected_count_lon = size // bins

        # Latitude plot
        observed_counts_lat, bins_lat_deg = np.histogram(lats, bins=bins, range=(-90, 90))
        # Convert bins to degrees
        bins_lon = np.deg2rad(bins_lon_deg)
        bins_lat = np.deg2rad(bins_lat_deg)
        
        # For expected counts in latitude, adjust for area covered by each bin
        area_ratio = np.sin(bins_lat[1:]) - np.sin(bins_lat[:-1])
        total_area = np.sin(np.pi/2) - np.sin(-np.pi/2)  # Total area for the entire sphere
        expected_count_lat = size * area_ratio / total_area

        # Bar width in degrees
        bar_width_lon = np.diff(bins_lon_deg)
        bar_width_lat = np.diff(bins_lat_deg)

        # Plotting
        fig, axs = plt.subplots(2, 1, figsize=(8, 6))

        # Longitude plot
        axs[0].bar(bins_lon_deg[:-1], observed_counts_lon, width=bar_width_lon, align='edge', label='Observed')
        axs[0].plot(bins_lon_deg[:-1], [expected_count_lon] * len(bins_lon_deg[:-1]), color='red', label='Expected')
        axs[0].set_xlabel('Longitude (deg)')
        axs[0].set_ylabel('Number')
        axs[0].legend()
        axs[0].set_title('Number vs Longitude')
        
        # Latitude
        axs[1].bar(bins_lat_deg[:-1], observed_counts_lat, width=bar_width_lat, align='edge', alpha=0.5, label='Observed')
        axs[1].plot(bins_lat_deg[:-1], expected_count_lat, label='Expected', color='red')
        axs[1].set_xlabel('Latitude (deg)')
        axs[1].set_ylabel('Number')
        axs[1].legend()
        axs[1].set_title('Number vs Latitude')

        plt.tight_layout()
        plt.show()


    def plot_random_angle():
        # Sample data generation
        size = 10000
        angles = get_random_impact_angle(size=size)
        

        # Number of bins
        bins = 50 
        observed_counts, bins_ang = np.histogram(angles, bins=bins, range=(0.0, 90.0))

        # Calculate expected distribution
        uniform_dist = np.linspace(0, 1, size)
        transformed_angles = np.rad2deg(np.arcsin(np.sqrt(uniform_dist)))
        expected_counts, _ = np.histogram(transformed_angles, bins=bins, range=(0.0, 90.0))

        # Plotting
        fig, ax = plt.subplots(figsize=(8, 4))

        # Observed counts
        ax.bar(bins_ang[:-1], observed_counts, width=np.diff(bins_ang), align='edge', label='Observed', alpha=0.5)

        # Expected counts
        ax.plot(bins_ang[:-1], expected_counts, label='Expected', color='red')

        ax.set_xlabel('Impact Angle (deg)')
        ax.set_ylabel('Count')
        ax.legend()
        ax.set_title('Impact Angle Distribution')

        plt.show()


    def plot_random_velocity():
        vmean = 20e3
        size = 10000
        velocities = get_random_velocity(vmean, size)
        
        # Create histogram of the generated velocities
        bins = np.linspace(0, vmean*3, 50)
        histogram, bins = np.histogram(velocities, bins=bins, density=True)
        bin_centers = 0.5 * (bins[1:] + bins[:-1])

        # Generate theoretical Maxwell-Boltzmann distribution for the given vmean
        theoretical_dist = maxwell.pdf(bin_centers, scale=vmean/np.sqrt(8/np.pi))

        # Plotting
        plt.figure(figsize=(8, 4))
        plt.bar(bin_centers, histogram, width=bins[1]-bins[0], alpha=0.5, label='Generated Velocities')
        plt.plot(bin_centers, theoretical_dist, label='Theoretical Maxwell-Boltzmann', color='red')
        plt.xlabel('Velocity (m/s)')
        plt.ylabel('Probability Density')
        plt.title('Maxwell-Boltzmann Distribution of Velocities')
        plt.legend()
        plt.show()

           
    def plot_random_size():
        num_realizations = 100
        nbins = 10
        size = 100000
        Dhi = 100.0
        p = 2.0
        C =  Dhi**p 
        Dlo = (size/C)**(-1.0/p) 
        diameters = np.exp(np.linspace(np.log(Dlo), np.log(100*Dhi), nbins))
        # Test a simple power law SFD
        cdf = C * diameters**-p
        
        fig = plt.figure()
        ax = fig.add_subplot(1, 1, 1)

        ax.set_title("Test SFD generation")
        ax.set_xlabel("$D$")
        ax.set_ylabel("$N_{>D}$")
        ax.set_xscale('log')
        ax.set_yscale('log')
        ax.set_ylim([1.0,1.1*size])

        for i in range(num_realizations):
            new_diameters = get_random_size(diameters, cdf, mu=size)
            new_diameters = np.sort(new_diameters)[::-1]
            nval = np.arange(1, new_diameters.size+1)
            if i == 0:
                label = 'Sampled SFD'
            else:
                label = None
                
            ax.plot(new_diameters, nval, label=label, alpha=0.25, color='orange')
            
        ax.plot(diameters, cdf, label='Model SFD')
        ax.legend(loc='upper right')
        plt.show()  
        
        
    plot_random_location()
    plot_random_angle()
    plot_random_velocity()
    plot_random_size()          


import unittest
import os
import shutil
import numpy as np
import tempfile
from cratermaker import Target
from cratermaker import Surface
from cratermaker.core.surface import IcosphereGrid, ArbitraryResolutionGrid, HiResLocalGrid, _DATA_DIR, _GRID_FILE_NAME
from cratermaker.utils.montecarlo import get_random_location
from cratermaker.utils.general_utils import normalize_coords

class TestSurface(unittest.TestCase):
    """
    A collection of unit tests for the Surface class in the cratermaker project.

    Parameters
    ----------
    temp_dir : TemporaryDirectory
        A temporary directory for testing file generation and I/O.
    grid_file : str
        Path to the temporary grid file.
    target : Target
        Target object representing a celestial body.
    pix : float
        Pixel size or resolution of the grid.

    """    
    def setUp(self):
        # Initialize a target and surface for testing
        self.temp_dir = tempfile.TemporaryDirectory()
        self.grid_file = os.path.join(self.temp_dir.name, _GRID_FILE_NAME)
        self.target = Target(name="Moon")
        self.pix = self.target.radius / 10.0
        self.gridlevel = 4
        os.chdir(self.temp_dir.name)
        
        return
    
    def tearDown(self):
        # Clean up temporary directory
        self.temp_dir.cleanup() 
        return

    def test_generate_grid(self):
        # Generate grid
        grid_strategy = IcosphereGrid(level=self.gridlevel, radius=self.target.radius)
        grid_strategy.generate_grid(grid_file=self.grid_file)
        self.assertTrue(os.path.exists(self.grid_file))
        
        grid_strategy = ArbitraryResolutionGrid(pix=self.pix, radius=self.target.radius) 
        grid_strategy.generate_grid(grid_file=self.grid_file)
        self.assertTrue(os.path.exists(self.grid_file))        
        
        grid_strategy = HiResLocalGrid(pix=self.pix, radius=self.target.radius, local_location=(0, 0), local_radius=100e3, superdomain_scale_factor=10)
        grid_strategy.generate_grid(grid_file=self.grid_file)
        self.assertTrue(os.path.exists(self.grid_file))
        return

    def test_initialize_surface(self):
        directory = os.path.join(os.getcwd(), _DATA_DIR)
        if os.path.exists(directory) and os.path.isdir(directory):
            try:
                shutil.rmtree(directory)
            except Exception as error:
                print(f"Error: {directory} : {error}")

        # Initializing it first should run the jigsaw mesh generator
        surf = Surface.initialize(gridlevel=self.gridlevel, target=self.target, reset_surface=True)
        
        # Try initializing the surface again with the same parameters. This should find the existing grid file and load it 
        surf = Surface.initialize(gridlevel=self.gridlevel, target=self.target, reset_surface=False)
        
        # Test regridding if the parameters change
        n_face_orig = surf.uxgrid.n_face
    
        surf = Surface.initialize(gridlevel=self.gridlevel-1, target=self.target, reset_surface=False)
        self.assertGreater(n_face_orig, surf.uxgrid.n_face)
    
        # Test different target values
        surf = Surface.initialize(gridlevel=self.gridlevel, target=Target(name="Mars"), reset_surface=False)
        surf = Surface.initialize(gridlevel=self.gridlevel, target="Mercury", reset_surface=False)
        
        # Test bad values
        with self.assertRaises(TypeError):
            surf = Surface.initialize(gridlevel=self.gridlevel, target=1, reset_surface=False)
        with self.assertRaises(ValueError):
            surf = Surface.initialize(gridlevel=self.gridlevel, target="Arrakis", reset_surface=False)
        with self.assertRaises(ValueError):
            surf = Surface.initialize(gridlevel=self.gridlevel, target=Target(name="Salusa Secundus"), reset_surface=False)
        return


    def test_set_elevation(self):
        surf = Surface.initialize(gridlevel=self.gridlevel, target=self.target, reset_surface=True)
        # Test with valid elevation data
        new_elev = np.random.rand(surf.uxgrid.n_node)  # Generate random elevation data
        surf.set_elevation(new_elev)

        # Test with invalid elevation data (wrong size)
        new_elev = np.random.rand(surf.uxgrid.n_node + 1)  # Incorrect size

        # Expect ValueError for incorrect size
        with self.assertRaises(ValueError):
            surf.set_elevation(new_elev)

        # Test setting elevation to None (should set to zero)
        surf.set_elevation(None)

        # Check if the elevation data is set to zero
        np.testing.assert_array_equal(surf['node_elevation'].values, np.zeros(surf.uxgrid.n_node))
        np.testing.assert_array_equal(surf['face_elevation'].values, np.zeros(surf.uxgrid.n_face))
        
        return
    
    
    def test_calculate_haversine_distance(self):
        # Example coordinates (lat/lon in radians)
        lon1, lat1 = np.radians(0), np.radians(0)  # Equator, prime meridian
        lon2, lat2 = np.radians(90), np.radians(0)  # 90 degrees East, equator
        radius = 6371000  # Earth's radius in meters

        # Known distance should be 1/4 the circumference of the Earth
        expected_distance = np.pi * radius / 2
        calculated_distance = Surface.calculate_haversine_distance(lon1, lat1, lon2, lat2, radius)

        # Compare the expected and calculated distances
        self.assertAlmostEqual(calculated_distance, expected_distance, places=1)


    def test_get_face_distance(self):
        surf = Surface.initialize(gridlevel=self.gridlevel, target=self.target, reset_surface=True)
        
        location = get_random_location()
        lon = location[0]
        lat = location[1]

        # Convert important points to radians
        north = (lon, lat + 90)
        south = (lon, lat - 90)
        antipode = (lon-180, -lat)
        
        north = normalize_coords(north)
        south = normalize_coords(south)
        antipode = normalize_coords(antipode)
        
        _,north_idx = surf.find_nearest_index(north)
        _,south_idx = surf.find_nearest_index(south)
        _,antipode_idx = surf.find_nearest_index(antipode)
        
        north_distance = surf.target.radius * np.pi / 2 
        south_distance = surf.target.radius * np.pi / 2
        antipode_distance = surf.target.radius * np.pi
        delta = 2*self.pix

        # Test distances
        _,distances = surf.get_distance(location)
        self.assertAlmostEqual(distances[north_idx], north_distance, delta=delta, msg=f"North face distance ratio: {distances[north_idx].item()/north_distance}")
        self.assertAlmostEqual(distances[south_idx], south_distance, delta=delta, msg=f"South face distance ratio: {distances[south_idx].item()/south_distance}")
        self.assertAlmostEqual(distances[antipode_idx], antipode_distance, delta=delta, msg=f"Antipode face distance ratio: {distances[antipode_idx].item()/antipode_distance}")
        
        
    def test_get_node_distance(self):
        surf = Surface.initialize(gridlevel=self.gridlevel, target=self.target, reset_surface=True)
        
        location = get_random_location()
        lon = location[0]
        lat = location[1]

        # Convert important points to radians
        north = (lon, lat + 90)
        south = (lon, lat - 90)
        antipode = (lon+180, -lat)
        
        north = normalize_coords(north)
        south = normalize_coords(south)
        antipode = normalize_coords(antipode)
        
        north_idx,_ = surf.find_nearest_index(north)
        south_idx,_ = surf.find_nearest_index(south)
        antipode_idx,_ = surf.find_nearest_index(antipode)

        north_distance = surf.target.radius * np.pi / 2 
        south_distance = surf.target.radius * np.pi / 2
        antipode_distance = surf.target.radius * np.pi
        delta = 2*self.pix

        # Test distances
        node_distances,_ = surf.get_distance(location)
        self.assertAlmostEqual(node_distances[north_idx], north_distance, delta=delta, msg=f"North node distance ratio: {node_distances[north_idx].item()/north_distance}")
        self.assertAlmostEqual(node_distances[south_idx], south_distance, delta=delta, msg=f"South node distance ratio: {node_distances[south_idx].item()/south_distance}")
        self.assertAlmostEqual(node_distances[antipode_idx], antipode_distance, delta=delta, msg=f"Antipode node distance ratio: {node_distances[antipode_idx].item()/antipode_distance}")
        
        
    def test_calculate_initial_bearing(self):
        # Example coordinates (lat/lon in radians)
        lon1, lat1 = np.radians(0), np.radians(0)  # Equator, prime meridian
        lon2, lat2 = np.radians(90), np.radians(0)  # 90 degrees East, equator

        # The bearing from (0, 0) to (90, 0) should be 90 degrees in radians
        expected_bearing = np.radians(90)
        calculated_bearing = Surface.calculate_initial_bearing(lon1, lat1, lon2, lat2)

        # Compare the expected and calculated bearings
        self.assertAlmostEqual(calculated_bearing, expected_bearing, places=1)
        
    # def test_get_random_on_face(self):
    #     # Tests that the random location is within the face we expect
    #     surf = Surface.initialize(gridlevel=self.gridlevel*2, target=self.target, reset_surface=True)
    #     n_per_face = 10
    #     for i in surf.n_face:
    #         original_face_index = i.values.item()
    #         for _ in range(n_per_face):
    #             location = surf.get_random_location_on_face(original_face_index)
    #             _, new_face_index = surf.find_nearest_index(location)
    #             self.assertEqual(original_face_index, new_face_index) 
            
    def test_face_surface_values(self):
        # Tests that the face_surface generates the correct values
        surf = Surface.initialize(gridlevel=self.gridlevel, target=self.target, reset_surface=True) 
        total_area_1 = surf.uxgrid.calculate_total_face_area()
        total_area_2 = surf.face_areas.sum().item()
        ratio = np.sqrt(total_area_2/total_area_1) / self.target.radius
        self.assertAlmostEqual(ratio, 1.0, places=2)
        self.assertAlmostEqual(total_area_2/ (4*np.pi*self.target.radius**2), 1.0, places=2)

if __name__ == '__main__':
     
    unittest.main()


import numpy as np
from numpy.random import Generator
from scipy.optimize import fsolve
from numpy.typing import ArrayLike
from typing import Any, Tuple
from .target import Target
from ..utils.custom_types import FloatLike
from .surface import Surface
from .target import Target
from .. import crater, ejecta
import json
import math
from scipy import fft
import os
from numpy.typing import NDArray

class Morphology:
    """
    An operations class for computing the morphology of a crater based on its size and target properties.

    This class encapsulates the logic for altering the topography of the surface based on the crater properties.

    Parameters
    ----------
    crater : Crater
        The crater to be created.
    target : Target
        The target body for the impact simulation.
    ejecta_truncation : float, optional
        The relative distance from the rim of the crater to truncate the ejecta blanket, default is None, which will compute a 
        truncation distance based on where the ejecta thickness reaches a small value.         
    rng : Generator, optional
        A random number generator instance. If not provided, the default numpy RNG will be used.
    dorays : bool, optional
        A flag to determine if the ray pattern should be used instead of the homogeneous ejecta blanket, default is True.
    **kwargs : Any
        Additional keyword arguments to be passed to internal functions.
    """
    
    def __init__(self, 
                 crater,  
                 target: Target | None = None, 
                 ejecta_truncation: FloatLike | None = None,
                 rng: Generator | None = None,
                 dorays: bool = True,
                 **kwargs: Any 
                 ):

        self.crater = crater 
        self.target = target
        self.rng = rng
        self.ejecta_truncation = ejecta_truncation
        self.dorays = dorays

        # Set the morphology based on crater type
        self._set_morphology_parameters()
       
    def __repr__(self):
        return (f"Morphology(morphology_type={self.morphology_type}, diameter={self.diameter}, "
                f"rimheight: {self.rimheight}, rimwidth: {self.rimwidth}, floordepth: {self.floordepth}, floordiam: {self.floordiam})") 
    
    def _set_morphology_parameters(self):
        """
        Sets the morphology parameters based on the crater type.
        """
        diameter_km = self.diameter * 1e-3  # Convert to km for these models

        if self.morphology_type in ["simple", "transitional"]:
            # A hybrid model between Pike (1977) and Fassett & Thomson (2014)
            self.rimheight = 0.043 * diameter_km**1.014 * 1e3  # Closer to Fassett & Thomson
            self.rimwidth = 0.257 * diameter_km**1.011 * 1e3   # Pike model
            self.floordepth = 0.224 * diameter_km**1.010 * 1e3 # Closer to Fassett & Thomson
            self.floordiam = 0.200 * diameter_km**1.143 * 1e3  # Fassett & Thomson for D~1km, Pike for D~20km

        elif self.morphology_type in ["complex", "peakring", "multiring"]:
            # Following Pike (1977)
            self.rimheight = 0.236 * diameter_km**0.399 * 1e3  # Pike model
            self.rimwidth = 0.467 * diameter_km**0.836 * 1e3   # Pike model
            self.floordepth = 1.044 * diameter_km**0.301 * 1e3 # Pike model
            # Fassett & Thomson for D~1km, Pike for D~20km, but limited to 90% of diameter
            self.floordiam = min(0.187 * diameter_km**1.249 * 1e3, 0.9 * self.diameter)
            self.peakheight = 0.032 * diameter_km**0.900 * 1e3  # Pike model
        else:
            raise ValueError(f"Unknown morphology type: {self.morphology_type}")
            
        self.ejrim = 0.14 * (self.diameter * 0.5)**(0.74) # McGetchin et al. (1973) Thickness of ejecta at rim


    def crater_profile(self, r: ArrayLike, r_ref: ArrayLike) -> np.float64:
        elevation = crater.profile(r,
                                   r_ref, 
                                   self.diameter, 
                                   self.floordepth, 
                                   self.floordiam, 
                                   self.rimheight, 
                                   self.ejrim
                                )
        
        return np.array(elevation, dtype=np.float64)
    

    def ejecta_profile(self, r: ArrayLike) -> np.float64:
        elevation = ejecta.profile(r,
                                   self.diameter, 
                                   self.ejrim
                                )
        elevation = np.array(elevation, dtype=np.float64)
        return elevation
   
    
    def ejecta_distribution(self, r: ArrayLike, theta: ArrayLike) -> np.float64:
        thickness = ejecta.distribution(r, theta,
                                       self.diameter, 
                                       self.ejrim, 
                                       self.ejecta_truncation,
                                       self.dorays
                                    )
        thickness = np.array(thickness, dtype=np.float64)
        return thickness
  
    def ray_intensity(self, r: ArrayLike, theta: ArrayLike) -> np.float64:
        intensity = ejecta.ray_intensity(r, theta,
                                       self.diameter, 
                                       self.ejrim, 
                                       self.ejecta_truncation,
                                    )
        intensity = np.array(intensity, dtype=np.float64)
        return intensity
           
    def compute_rmax(self, 
                     minimum_thickness: np.float64,
                     feature: str = "ejecta") -> np.float64:
        """
        Compute the maximum extent of the crater based on the minimum thickness of a feature, or the ejecta_truncation factor,
        whichever is smaller.
        
        Parameters
        ----------
        minimum_thickness : np.float64
            The minimum thickness of the feature blanket in meters.
        feature : str, optional, default = "ejecta"
            The feature to compute the maximum extent. Either "crater" or "ejecta". If "crater" is chosen, the rmax is based
            on where the raised rim is smaller than minimum thickness. 
        """ 
        
        # Compute the reference surface for the crater 

        if feature == "ejecta":
            def _profile_invert(r):
                return self.ejecta_profile(r) - minimum_thickness
        elif feature == "crater":
            def _profile_invert(r):
                return self.crater_profile(r, np.zeros(1)) - minimum_thickness
        else:
            raise ValueError("Unknown feature type. Choose either 'crater' or 'ejecta'")
    
        # Get the maximum extent 
        rmax = fsolve(_profile_invert, x0=self.radius*1.01)[0]
        
        if self.ejecta_truncation:
            rmax = min(rmax, self.ejecta_truncation * self.radius)
            
        if feature == "crater":
            rmax = max(rmax, self.radius)

        return rmax     
     
    def form_crater(self, 
                    surf: Surface,
                    **kwargs) -> None:
        """
        This method forms the interior of the crater by altering the elevation variable of the surface mesh.
        
        Parameters
        ----------
        surf : Surface
            The surface to be altered.
        **kwargs : dict
            Additional keyword arguments to be passed to internal functions (not used here).
        """
        
        # Test if the crater is big enough to modify the surface
        rmax = self.compute_rmax(minimum_thickness=surf.smallest_length)
        region_surf = surf.extract_region(self.crater.location, rmax)
        if not region_surf: # The crater is too small to change the surface
            return
        crater_area = np.pi * rmax**2
        
        # Check to make sure that the face at the crater location is not smaller than the crater area
        if surf['face_areas'].isel(n_face=self.crater.face_index) > crater_area:
            return
        
        region_surf['node_crater_distance'], region_surf['face_crater_distance'] = region_surf.get_distance(self.crater.location)
        region_surf.get_reference_surface(self.crater.location, self.crater.radius)
        
        try:
            node_elevation = self.crater_profile(region_surf['node_crater_distance'].values, 
                                                 region_surf['reference_node_elevation'].values)
            surf['node_elevation'].loc[{'n_node': region_surf.uxgrid._ds["subgrid_node_indices"]}] = node_elevation
            
            face_elevation = self.crater_profile(region_surf['face_crater_distance'].values, 
                                                 region_surf['reference_face_elevation'].values)
            surf['face_elevation'].loc[{'n_face': region_surf.uxgrid._ds["subgrid_face_indices"]}] = face_elevation
        except:
            print(self)
            raise ValueError("Something went wrong with this crater!")
                 
        return  
    
     
    def form_ejecta(self,
                    surf: Surface,
                    **kwargs) -> None:
        """
        This method forms the ejecta blanket around the crater by altering the elevation variable of the surface mesh.
       
        Parameters
        ----------
        surf : Surface
            The surface to be altered.
        **kwargs : dict
            Additional keyword arguments to be passed to internal functions (not used here). 
        """
                 
        # Test if the ejecta is big enough to modify the surface
        rmax = self.compute_rmax(minimum_thickness=surf.smallest_length) 
        if not self.ejecta_truncation:
            self.ejecta_truncation = rmax / self.radius
        region_surf = surf.extract_region(self.crater.location, rmax)
        if not region_surf: # The crater is too small to change the surface
            return
        ejecta_area = np.pi * rmax**2
        
        # Check to make sure that the face at the crater location is not smaller than the ejecta blanket area
        if surf['face_areas'].isel(n_face=self.crater.face_index) > ejecta_area:
            return                  
        
        region_surf['node_crater_distance'], region_surf['face_crater_distance'] = region_surf.get_distance(self.crater.location)
        region_surf['node_crater_bearing'], region_surf['face_crater_bearing']  = region_surf.get_initial_bearing(self.crater.location)
        
        try:
            node_thickness = self.ejecta_distribution(region_surf['node_crater_distance'].values, 
                                                     region_surf['node_crater_bearing'].values)
            surf['node_elevation'].loc[{'n_node': region_surf.uxgrid._ds["subgrid_node_indices"]}] += node_thickness
            
            face_thickness = self.ejecta_distribution(region_surf['face_crater_distance'].values, 
                                                     region_surf['face_crater_bearing'].values)
            surf['face_elevation'].loc[{'n_face': region_surf.uxgrid._ds["subgrid_face_indices"]}] += face_thickness
            surf['ejecta_thickness'].loc[{'n_face': region_surf.uxgrid._ds["subgrid_face_indices"]}] += face_thickness
            
            if self.dorays: 
                face_intensity = self.ray_intensity(region_surf['face_crater_distance'].values, 
                                                     region_surf['face_crater_bearing'].values)
                surf['ray_intensity'].loc[{'n_face': region_surf.uxgrid._ds["subgrid_face_indices"]}] += face_intensity
        except:
            print(self)
            raise ValueError("Something went wrong with this crater!")
                 
        return  
    
    def form_secondaries(self,
                         surf: Surface,
                         **kwargs) -> None:
        """
        This method forms secondary craters around the primary crater. Currently it only generates the ray intensity function.
        Maximum ray length formula is from [1]_.
       
        Parameters
        ----------
        surf : Surface
            The surface to be altered.
        **kwargs : dict
            Additional keyword arguments to be passed to internal functions (not used here). 
            
        References
        ----------
        .. [1] Elliott, J.R., Huang, Y.-H., Minton, D.A., Freed, A.M., 2018. The length of lunar crater rays explained using secondary crater scaling. Icarus 312, 231–246. https://doi.org/10.1016/j.icarus.2018.04.015

        """
        
        # Elliott et al. (2018) eq. 2
        A = 6.59 + self.rng.normal(loc=0.0,scale=1.45) 
        p = 1.27 + self.rng.normal(loc=0.0,scale=0.06)
        rmax = A * (self.radius / 1000) **p * 1000
        if not self.ejecta_truncation:
            self.ejecta_truncation = rmax / self.radius
        region_surf = surf.extract_region(self.crater.location, rmax)
        if not region_surf: # The crater is too small to change the surface
            return
        ray_area = np.pi * rmax**2
        
        # Check to make sure that the face at the crater location is not smaller than the ray effectt area
        if surf['face_areas'].isel(n_face=self.crater.face_index) > ray_area:
            return                  
        
        region_surf['node_crater_distance'], region_surf['face_crater_distance'] = region_surf.get_distance(self.crater.location)
        region_surf['node_crater_bearing'], region_surf['face_crater_bearing']  = region_surf.get_initial_bearing(self.crater.location)
        
        try:
            face_intensity = self.ray_intensity(region_surf['face_crater_distance'].values, 
                                                region_surf['face_crater_bearing'].values)
            surf['ray_intensity'].loc[{'n_face': region_surf.uxgrid._ds["subgrid_face_indices"]}] += face_intensity
        except:
            print(self)
            raise ValueError("Something went wrong with this crater!")
                 
        return      

    def get_1D_power_spectral_density(self,feature,psd_coef,num_psd_component_effec=100) -> NDArray:
        """
        This method construct a 1D power spectral density.
        Coeffcients are from [1]_.
       
        Parameters
        ----------
        feature : string
            For a 1D feature, choose from ejecta, rim, and floor
        psd_coef : dict (.json)
            Coeffcients used to constract a 1D power spectral density
        num_psd_component_effec : int
            Only reconstrut the sine waves with wavelengths smaller than 2pi/num_psd_component_effec to improve computational efficiency

        References
        ----------
        .. [1] Du, J., Minton, D. A., Blevins, A. M., Fassett, C. I., & Huang, Y. H. (2024). Spectral analysis of the morphology of fresh lunar craters I: Rim crest, floor, and rim flank outlines. Journal of Geophysical Research: Planets, 129(11), e2024JE008357. https://doi.org/10.1029/2024JE008357

        """
        # read the psd_coef outside this function. this is temporary until we find a better way to read the psd_coef
        with open(os.path.join(os.pardir,"plugins","psd_coef.json")) as f:
            psd_coef = json.load(f)
        # ------------------------------------------------------------------------------------------------------------------
        num_psd_component=5000
        # ------------------------------------------------------------------------------------------------------------------
        if self.crater.diameter<psd_coef["1D"][feature]["slope_12"]["D_tie"]:
            slope_12 = psd_coef["1D"][feature]["slope_12"]["k1"] * self.crater.diameter + psd_coef["1D"][feature]["slope_12"]["b1"]
            slope_12_sigma = psd_coef["1D"][feature]["slope_12"]["sigma_1"]
        else:
            slope_12 = psd_coef["1D"][feature]["slope_12"]["k2"] * self.crater.diameter + psd_coef["1D"][feature]["slope_12"]["b2"]
            slope_12_sigma = psd_coef["1D"][feature]["slope_12"]["sigma_2"]
        if self.crater.diameter<psd_coef["1D"][feature]["bp2_x"]["D_tie"]:
            bp2_x = psd_coef["1D"][feature]["bp2_x"]["k1"] * self.crater.diameter + psd_coef["1D"][feature]["bp2_x"]["b1"]
            bp2_x_sigma = psd_coef["1D"][feature]["bp2_x"]["sigma_1"]
        else:
            bp2_x = psd_coef["1D"][feature]["bp2_x"]["k2"] * self.crater.diameter + psd_coef["1D"][feature]["bp2_x"]["b2"]
            bp2_x_sigma = psd_coef["1D"][feature]["bp2_x"]["sigma_2"]
        if self.crater.diameter<psd_coef["1D"][feature]["bp2_y"]["D_tie"]:
            bp2_y = psd_coef["1D"][feature]["bp2_y"]["k1"] * self.crater.diameter + psd_coef["1D"][feature]["bp2_y"]["b1"]
            bp2_y_sigma = psd_coef["1D"][feature]["bp2_y"]["sigma_1"]
        else:
            bp2_y = psd_coef["1D"][feature]["bp2_y"]["k2"] * self.crater.diameter + psd_coef["1D"][feature]["bp2_y"]["b2"]
            bp2_y_sigma = psd_coef["1D"][feature]["bp2_y"]["sigma_2"]
        if self.crater.diameter<psd_coef["1D"][feature]["bp3_y"]["D_tie"]:
            bp3_y = psd_coef["1D"][feature]["bp3_y"]["k1"] * self.crater.diameter + psd_coef["1D"][feature]["bp3_y"]["b1"]
            bp3_y_sigma = psd_coef["1D"][feature]["bp3_y"]["sigma_1"]
        else:
            bp3_y = psd_coef["1D"][feature]["bp3_y"]["k2"] * self.crater.diameter + psd_coef["1D"][feature]["bp3_y"]["b2"]
            bp3_y_sigma = psd_coef["1D"][feature]["bp3_y"]["sigma_2"]
        if self.crater.diameter<psd_coef["1D"][feature]["bp4_y"]["D_tie"]:
            bp4_y = psd_coef["1D"][feature]["bp4_y"]["k1"] * self.crater.diameter + psd_coef["1D"][feature]["bp4_y"]["b1"]
            bp4_y_sigma = psd_coef["1D"][feature]["bp4_y"]["sigma_1"]
        else:
            bp4_y = psd_coef["1D"][feature]["bp4_y"]["k2"] * self.crater.diameter + psd_coef["1D"][feature]["bp4_y"]["b2"]
            bp4_y_sigma = psd_coef["1D"][feature]["bp4_y"]["sigma_2"]
        slope_12 = np.random.normal(slope_12, slope_12_sigma)
        bp2_x = np.random.normal(bp2_x, bp2_x_sigma)
        bp2_y = np.random.normal(bp2_y, bp2_y_sigma)
        bp3_y = np.random.normal(bp3_y, bp3_y_sigma)
        bp4_y = np.random.normal(bp4_y, bp4_y_sigma)
        psd_sigma=psd_coef["1D"][feature]["psd_sigma"]
        # ------------------------------------------------------------------------------------------------------------------
        bp4_x = math.log10(2*math.pi)
        bp3_x =  math.log10(10**bp4_x / 2)
        interval=2*math.pi/num_psd_component
        freq = fft.fftfreq(num_psd_component, interval)
        wavelength = 1 / freq[1:num_psd_component//2]
        psd = np.zeros((num_psd_component//2 - 1, 2))
        psd[:, 0] = wavelength
        # ----------------------------------------------------------------------------------------------------------------------
        bp2_x_index = int(  10**bp4_x/10**bp2_x-1)
        k_23=(bp3_y-bp2_y)/(bp3_x-bp2_x)
        b_23 = bp3_y-k_23*bp3_x
        k_12=slope_12
        b_12 = bp2_y-k_12*bp2_x
        #-----------------------------------------------------------------------------------------------------------------------
        psd[0,1] =10**bp4_x
        psd[1,1]=10**bp3_y
        psd[2:bp2_x_index+1,1] =10**(k_23*np.log10(psd[2:bp2_x_index+1, 0])+b_23)
        psd[bp2_x_index+1:, 1] =10**(k_12*np.log10(psd[bp2_x_index+1:, 0])+b_12)
        # ----------------------------------------------------------------------------------------------------------------------
        psd=psd[:num_psd_component_effec]
        psd_log = np.log10(psd[:, 1])
        psd_log += np.random.normal(0, psd_sigma, psd_log.shape) 
        psd[:, 1] = 10 ** psd_log
        psd=np.flipud(psd)
        return psd
            
    def get_2D_power_spectral_density(self,feature,psd_coef,ejecta_radius_norm,floor_radius_norm,max_effec_freq) -> Tuple[NDArray,NDArray,NDArray,NDArray,NDArray,NDArray]:
        """
        This method constructs a 2D power spectral density.
        Coeffcients are from [1]_.
       
        Parameters
        ----------
        feature : string
            For a 2D feature, choose from ejecta, wall, and floor
        psd_coef : dict (.json)
            Coeffcients used to constract a 2D power spectral density
        ejecta_radius_norm : float
            The radius of the continuous ejecta normalized by the crater radius
        floor_radius_norm : float
            The radius of the floor normalized by the crater radius
        max_effec_freq : int
            Only reconstrut the sine waves with frequencies smaller than a maxmium effective frequency to improve computational efficiency 

        References
        ----------
        .. [1] Du, J., Minton, D. A., Blevins, A. M., Fassett, C. I., & Huang, Y. H. (2025). Spectral Analysis of the Morphology of Fresh Lunar Craters II: Two-Dimensional Surface Elevations of the Continuous Ejecta, Wall, and Floor. Journal of Geophysical Research: Planets

        """
         # read the psd_coef outside this function. this is temporary until we find a better way to read the psd_coef
        with open(os.path.join(os.pardir,"plugins","psd_coef.json")) as f:
            psd_coef = json.load(f) # read in from _init_?
        # ------------------------------------------------------------------------------------------------------------------
        if feature == "ejecta" : max_effec_freq = 40
        if feature == "wall" : max_effec_freq = 40
        if feature == "floor" : max_effec_freq = 100
        # ------------------------------------------------------------------------------------------------------------------
        if feature=="ejecta":
            theta_min = 0
            theta_max = 2 * math.pi
            r_min=1
            r_max=ejecta_radius_norm
            area=(r_max - r_min) * (theta_max - theta_min)
            theta_number = 2000
            r_number = 400
        if feature == "wall":
            theta_min = 0
            theta_max = 2 * math.pi
            r_min = floor_radius_norm
            r_max = 1
            area= (r_max - r_min) * (theta_max - theta_min)
            theta_number = 1000
            r_number = 200
        if feature == "floor":
            r_min = -floor_radius_norm/math.sqrt(2)
            r_max = floor_radius_norm/math.sqrt(2)
            theta_min = r_min
            theta_max = r_max
            area=(r_max - r_min) ** 2
            theta_number = 512
            r_number = 512
        # ------------------------------------------------------------------------------------------------------------------
        side_r=np.linspace(0, r_max-r_min, num=r_number, endpoint=False)
        side_theta=np.linspace(0, theta_max-theta_min, num=theta_number, endpoint=False)
        grid_theta, grid_r = np.meshgrid( side_theta,side_r)
        side_freq_theta = np.fft.fftfreq(theta_number, side_theta[1] - side_theta[0])
        side_freq_r = np.fft.fftfreq(r_number, side_r[1] - side_r[0])
        freq_theta, freq_r = np.meshgrid(np.fft.fftshift(side_freq_theta), np.fft.fftshift(side_freq_r))
        freq_theta_quadrant = freq_theta [r_number // 2:, theta_number // 2:]
        freq_r_quadrant = freq_r  [r_number // 2:, theta_number // 2:]
        # ------------------------------------------------------------------------------------------------------------------
        if self.crater.diameter<psd_coef["2D"][feature]["p_max"]["D_tie"]:
            p_max = psd_coef["2D"][feature]["p_max"]["k1"] * self.crater.diameter + psd_coef["2D"][feature]["p_max"]["b1"]
            p_max_sigma = psd_coef["2D"][feature]["p_max"]["sigma_1"]
        else:
            p_max = psd_coef["2D"][feature]["p_max"]["k2"] * self.crater.diameter + psd_coef["2D"][feature]["p_max"]["b2"]
            p_max_sigma = psd_coef["2D"][feature]["p_max"]["sigma_2"]
        if self.crater.diameter<psd_coef["2D"][feature]["p_diff"]["D_tie"]:
            p_diff = psd_coef["2D"][feature]["p_diff"]["k1"] * self.crater.diameter + psd_coef["2D"][feature]["p_diff"]["b1"]
            p_diff_sigma = psd_coef["2D"][feature]["p_diff"]["sigma_1"]
        else:
            p_diff = psd_coef["2D"][feature]["p_diff"]["k2"] * self.crater.diameter + psd_coef["2D"][feature]["p_diff"]["b2"]
            p_diff_sigma = psd_coef["2D"][feature]["p_diff"]["sigma_2"]
        if self.crater.diameter<psd_coef["2D"][feature]["nu_fall"]["D_tie"]:
            nu_fall = psd_coef["2D"][feature]["nu_fall"]["k1"] * self.crater.diameter + psd_coef["2D"][feature]["nu_fall"]["b1"]
            nu_fall_sigma = psd_coef["2D"][feature]["nu_fall"]["sigma_1"]
        else:
            nu_fall = psd_coef["2D"][feature]["nu_fall"]["k2"] * self.crater.diameter + psd_coef["2D"][feature]["nu_fall"]["b2"]
            nu_fall_sigma = psd_coef["2D"][feature]["nu_fall"]["sigma_2"]
        if self.crater.diameter<psd_coef["2D"][feature]["psd_sigma"]["D_tie"]:
            psd_sigma = psd_coef["2D"][feature]["psd_sigma"]["k1"] * self.crater.diameter + psd_coef["2D"][feature]["psd_sigma"]["b1"]
            psd_sigma_sigma = psd_coef["2D"][feature]["psd_sigma"]["sigma_1"]
        else:
            psd_sigma = psd_coef["2D"][feature]["psd_sigma"]["k2"] * self.crater.diameter + psd_coef["2D"][feature]["psd_sigma"]["b2"]
            psd_sigma_sigma = psd_coef["2D"][feature]["psd_sigma"]["sigma_2"]
        p_max = np.random.normal(p_max, p_max_sigma)
        p_diff = np.random.normal(p_diff, p_diff_sigma)
        nu_fall = np.random.normal(nu_fall, nu_fall_sigma)
        psd_sigma = np.random.normal(psd_sigma, psd_sigma_sigma)
        # ------------------------------------------------------------------------------------------------------------------
        freq_r_matrix= np.sqrt(freq_theta_quadrant**2+freq_r_quadrant**2)
        freq_theta_matrix = np.arctan2(freq_r_quadrant,freq_theta_quadrant)
        if feature=="ejecta":
            if self.crater.diameter<psd_coef["2D"][feature]["E_rad"]["D_tie"]:
                E_rad = psd_coef["2D"][feature]["E_rad"]["k1"] * self.crater.diameter + psd_coef["2D"][feature]["E_rad"]["b1"]
                E_rad_sigma = psd_coef["2D"][feature]["E_rad"]["sigma_1"]
            else:
                E_rad = psd_coef["2D"][feature]["E_rad"]["k2"] * self.crater.diameter + psd_coef["2D"][feature]["E_rad"]["b2"]
                E_rad_sigma = psd_coef["2D"][feature]["E_rad"]["sigma_2"]
            E_rad = np.random.normal(E_rad, E_rad_sigma)
            psd_log_quadrant=p_diff/np.sqrt(1-(E_rad*np.sin(freq_theta_matrix))**2)*(np.exp(-np.sqrt(freq_r_matrix/nu_fall))-1)+p_max
        if feature=="wall":
            if self.crater.diameter<psd_coef["2D"][feature]["E_circ"]["D_tie"]:
                E_circ = psd_coef["2D"][feature]["E_circ"]["k1"] * self.crater.diameter + psd_coef["2D"][feature]["E_circ"]["b1"]
                E_circ_sigma = psd_coef["2D"][feature]["E_circ"]["sigma_1"]
            else:
                E_circ = psd_coef["2D"][feature]["E_circ"]["k2"] * self.crater.diameter + psd_coef["2D"][feature]["E_circ"]["b2"]
                E_circ_sigma = psd_coef["2D"][feature]["E_circ"]["sigma_2"]
            E_circ = np.random.normal(E_circ, E_circ_sigma)
            psd_log_quadrant=p_diff/np.sqrt(1-(E_circ*np.cos(freq_theta_matrix))**2)*(np.exp(-np.sqrt(freq_r_matrix/nu_fall))-1)+p_max
        if feature=="floor":
            psd_log_quadrant = p_diff  * (np.exp(-np.sqrt(freq_r_matrix / nu_fall)) - 1) + p_max
        psd_log_quadrant_1st=np.random.normal(psd_log_quadrant,psd_sigma)
        psd_log_quadrant_2nd = np.random.normal(psd_log_quadrant, psd_sigma)
        # ------------------------------------------------------------------------------------------------------------------
        psd_log = np.zeros((r_number, theta_number))
        psd_log[r_number//2:, theta_number // 2:] =  psd_log_quadrant_1st
        psd_log[r_number//2+1:, 1:theta_number // 2] = np.fliplr(psd_log_quadrant_2nd[1:,1:])
        psd_log[ 1:r_number // 2, theta_number // 2+1:] = np.flipud(psd_log_quadrant_2nd[1:, 1:])
        psd_log[1:r_number // 2, 1:theta_number // 2] = np.fliplr(np.flipud(psd_log_quadrant_1st[1:, 1:]))
        #---------------------------------------------------------------------------------------------------------------
        psd_log[1:r_number // 2, theta_number // 2] = np.flip(psd_log[r_number // 2+1:, theta_number // 2])
        psd_log[r_number // 2, 1:theta_number // 2] = np.flip(  psd_log[r_number // 2, theta_number // 2+1:] )
        #---------------------------------------------------------------------------------------------------------------
        psd_log[0,theta_number//2+1:]=psd_log[1,theta_number//2+1:]
        psd_log[r_number // 2 + 1:, 0] = psd_log[r_number // 2 + 1:, 1]
        #---------------------------------------------------------------------------------------------------------------
        psd_log[0, 1:theta_number // 2] = np.flip(psd_log[0,theta_number//2+1:])
        psd_log[1:r_number // 2:, 0] = np.flip(psd_log[r_number // 2 + 1:,0] )
        # --------------------------------------------------------------------------------------------------------------
        psd_log[0, 0]=psd_log[1, 1]
        psd_log[0, theta_number//2] =psd_log[1, theta_number//2]
        psd_log[r_number//2, 0] = psd_log[r_number//2, 1]
        # ------------------------------------------------------------------------------------------------------------------
        psd = 10 ** psd_log
        phase = np.random.uniform(-math.pi, math.pi, size=(r_number, theta_number))
        if feature=="ejecta" or feature=="wall" : psd[:,theta_number//2]=np.zeros((r_number))
        # ------------------------------------------------------------------------------------------------------------------
        freq_theta_delta=(freq_theta.max()-freq_theta.min())/theta_number
        freq_r_delta = (freq_r.max() - freq_r.min()) / r_number
        freq_theta_num=int(max_effec_freq/freq_theta_delta)
        freq_r_num = int(max_effec_freq / freq_r_delta)
        freq_theta = freq_theta[r_number//2- freq_r_num:r_number//2+ freq_r_num,theta_number//2-freq_theta_num:theta_number//2+ freq_theta_num]
        freq_r = freq_r[r_number//2- freq_r_num:r_number//2+ freq_r_num,theta_number//2-freq_theta_num:theta_number//2+ freq_theta_num]
        psd = psd[r_number//2- freq_r_num:r_number//2+ freq_r_num,theta_number//2-freq_theta_num:theta_number//2+ freq_theta_num]
        phase = phase[r_number//2- freq_r_num:r_number//2+ freq_r_num,theta_number//2-freq_theta_num:theta_number//2+ freq_theta_num]
        # ------------------------------------------------------------------------------------------------------------------
        psd=np.sqrt(psd*area)
        return psd,phase,freq_theta,freq_r,grid_theta,grid_r  

    @property
    def diameter(self) -> np.float64:
        """
        The diameter of the crater in meters.
        
        Returns
        -------
        np.float64
        """
        return self.crater.diameter

    @property
    def radius(self) -> np.float64:
        """
        The radius of the crater in meters.
        
        Returns
        -------
        np.float64
        """
        return self.crater.radius
   
    @property 
    def morphology_type(self) -> str:
        """
        The morphology type of the crater.
        
        Returns
        -------
        str 
        """
        return self.crater.morphology_type 
    
    @property
    def rimheight(self) -> np.float64:
        """
        The height of the crater rim in meters.
        
        Returns
        -------
        np.float64
        """
        return self.crater.rimheight
    
    @rimheight.setter
    def rimheight(self, value: FloatLike) -> None:
        if not isinstance(value, FloatLike):
            raise TypeError("rimheight must be of type FloatLike")
        self.crater.rimheight = np.float64(value)
        
    @property
    def rimwidth(self) -> np.float64:
        """
        The width of the crater rim in meters.
        
        Returns
        -------
        np.float64
        """
        return self.crater.rimwidth
    
    @rimwidth.setter
    def rimwidth(self, value: FloatLike) -> None:
        if not isinstance(value, FloatLike):
            raise TypeError("rimwidth must be of type FloatLike") 
        self.crater.rimwidth = np.float64(value)
        
    @property
    def peakheight(self) -> np.float64:
        """
        The height of the central peak in meters.
        
        Returns
        -------
        np.float64
        """
        return self.crater.peakheight
    
    @peakheight.setter
    def peakheight(self, value: FloatLike) -> None:
        if not isinstance(value, FloatLike):
            raise TypeError("peakheight must be of type FloatLike") 
        self.crater.peakheight = np.float64(value)

    @property
    def floordiam(self) -> np.float64:
        """
        The diameter of the crater floor in meters.
        
        Returns
        -------
        np.float64
        """
        return self.crater.floordiam
    
    @floordiam.setter
    def floordiam(self, value: FloatLike) -> None:
        if not isinstance(value, FloatLike):
            raise TypeError("floordiam must be of type FloatLike")
        self.crater.floordiam = np.float64(value)
        
    @property
    def floordepth(self) -> np.float64:
        """
        Return the depth of the crater floor in m
        
        Returns
        -------
        np.float64
        """
        return self.crater.floordepth
    
    @floordepth.setter
    def floordepth(self, value: FloatLike) -> None:
        if not isinstance(value, FloatLike):
            raise TypeError("floordepth must be of type FloatLike")
        self.crater.floordepth = np.float64(value)
        
    @property
    def ejrim(self) -> np.float64:
        """
        The thickness of ejecta at the rim in m.
        
        Returns
        -------
        np.float64
        """
        return self.crater.ejrim
    
    @ejrim.setter
    def ejrim(self, value: FloatLike) -> None:
        if not isinstance(value, FloatLike):
            raise TypeError("ejrim must be of type FloatLike")
        self.crater.ejrim = np.float64(value)
        
    @property
    def crater(self):
        """
        The crater to be created.
        
        Returns
        -------
        Crater
        """ 
        return self._crater
    
    @crater.setter
    def crater(self, value):
        from .impact import Crater
        if value is not None and not isinstance(value, Crater):
            raise TypeError("crater must be an instance of Crater")
        self._crater = value
        return 
        
    @property
    def target(self):
        """
        The target body for the impact.
        
        Returns
        -------
        Target
        """ 
        return self._target
    
    @target.setter
    def target(self, value):
        from .target import Target
        if value is None:
            self._target = Target(name="Moon")
            return
        if not isinstance(value, Target):
            raise TypeError("target must be an instance of Target")
        self._target = value
        return        
        
    @property
    def rng(self):
        """
        A random number generator instance.
        
        Returns
        -------
        Generator
        """ 
        return self._rng
    
    @rng.setter
    def rng(self, value):
        if not isinstance(value, Generator) and value is not None:
            raise TypeError("The 'rng' argument must be a numpy.random.Generator instance or None")
        self._rng = value or np.random.default_rng()   
       
    @property
    def ejecta_truncation(self) -> np.float64:
        """
        The radius at which the crater is truncated relative to the crater radius.
        
        Returns
        -------
        np.float64 or None
        """
        return self.crater.ejecta_truncation 
    
    @ejecta_truncation.setter
    def ejecta_truncation(self, value: FloatLike | None): 
        if value:
            if not isinstance(value, FloatLike):
                raise TypeError("truction_radius must be of type FloatLike")
            self.crater.ejecta_truncation = np.float64(value)
        else:
            self.crater.ejecta_truncation = None
            
    @property
    def dorays(self) -> bool:
        """
        A flag to determine if the ray pattern should be used instead of the homogeneous ejecta blanket.
        
        Returns
        -------
        bool
        """
        return self._dorays
    
    @dorays.setter
    def dorays(self, value: bool) -> None:
        if not isinstance(value, bool):
            raise TypeError("dorays must be of type bool")
        self._dorays = value
        return


# cython: language_level=3, c_string_type=unicode, c_string_encoding=ascii
cimport cython
cimport numpy as cnp
import numpy as np
from libc.stdlib cimport malloc, free

cdef extern from "ejecta.h":
    void bind_ejecta_distribution(double *radial_distance, double *initial_bearing, int num_elements, double crater_diameter, double ejrim, double ejecta_truncation, int dorays, double *ejecta_thickness);
    void bind_ejecta_profile(double *radial_distance, int num_elements, double crater_diameter, double ejrim, double *elevation)
    void bind_ejecta_ray_intensity(double *radial_distance, double *initial_bearing, int num_elements, double crater_diameter, double ejrim, double ejecta_truncation, double *ejecta_thickness);


def distribution(cnp.ndarray[cnp.float64_t, ndim=1] radial_distance, 
                cnp.ndarray[cnp.float64_t, ndim=1] initial_bearing,
                cnp.float64_t crater_diameter,
                cnp.float64_t ejrim,
                cnp.float64_t ejecta_truncation,
                bint dorays):
    """
    Generate an ejecta profile.

    Parameters
    ----------
    radial_distance : ndarray(N,)
        Radial distance from the center of the crater.
    initial_bearing : ndarray(N,)
        Initial bearing from the denter of the crater in radians.
    crater_diameter : float
        Diameter of the crater.
    ejrim : float
        Ejecta rim height.
    ejecta_truncation : float
        Ejecta truncation distance relative to the crater radius.
    dorays : bool
        Set to True to compute a ray pattern. Set to False to compute a homogenous distribution. 

    Returns
    -------
    ndarray(N,)
        computed elevation values

    Raises
    ------
    ValueError - If required parameters are missing, arrays are mismatched, or if an invalid model is specified.
    """

    cdef int num_elements = radial_distance.size

    # Ensure memory-contiguous numpy array
    radial_distance = np.ascontiguousarray(radial_distance, dtype=np.float64)
    initial_bearing = np.ascontiguousarray(initial_bearing, dtype=np.float64)
    thickness_array = np.empty(num_elements, dtype=np.float64)

    # Make memory view of the numpy arrays
    cdef cnp.float64_t[::1] r = radial_distance
    cdef cnp.float64_t[::1] theta = initial_bearing
    cdef cnp.float64_t[::1] ejecta_thickness = thickness_array
    cdef int c_dorays = 1 if dorays else 0

    bind_ejecta_distribution(&r[0], &theta[0], num_elements, crater_diameter, ejrim, ejecta_truncation, c_dorays, &ejecta_thickness[0])

    return ejecta_thickness  


def profile(cnp.ndarray[cnp.float64_t, ndim=1] radial_distance, 
            cnp.float64_t crater_diameter,
            cnp.float64_t ejrim):
    """
    Generate an ejecta profile.

    Parameters
    ----------
    radial_distance : ndarray(N,)
        radial distance from the center of the crater.
    crater_diameter : float
        diameter of the crater.
    ejrim : float
        ejecta rim height.

    Returns
    -------
    ndarray(N,)
        computed elevation values

    Raises
    ------
    ValueError - If required parameters are missing, arrays are mismatched, or if an invalid model is specified.
    """

    cdef int num_elements = radial_distance.size

    # Ensure memory-contiguous numpy array
    radial_distance = np.ascontiguousarray(radial_distance, dtype=np.float64)
    elevation_array = np.empty(num_elements, dtype=np.float64)

    # Make memory view of the numpy arrays
    cdef cnp.float64_t[::1] r = radial_distance
    cdef cnp.float64_t[::1] elevation = elevation_array

    bind_ejecta_profile(&r[0], num_elements, crater_diameter, ejrim, &elevation[0])

    return elevation  

def ray_intensity(cnp.ndarray[cnp.float64_t, ndim=1] radial_distance, 
                cnp.ndarray[cnp.float64_t, ndim=1] initial_bearing,
                cnp.float64_t crater_diameter,
                cnp.float64_t ejrim,
                cnp.float64_t ejecta_truncation):
    """
    Generate an ejecta profile.

    Parameters
    ----------
    radial_distance : ndarray(N,)
        radial distance from the center of the crater.
    initial_bearing : ndarray(N,)
        initial bearing from the denter of the crater in radians.
    crater_diameter : float
        diameter of the crater.
    ejrim : float
        ejecta rim height.
    ejecta_truncation : float
        ejecta truncation distance relative to the crater radius.

    Returns
    -------
    ndarray(N,)
        computed elevation values

    Raises
    ------
    ValueError - If required parameters are missing, arrays are mismatched, or if an invalid model is specified.
    """

    cdef int num_elements = radial_distance.size

    # Ensure memory-contiguous numpy array
    radial_distance = np.ascontiguousarray(radial_distance, dtype=np.float64)
    initial_bearing = np.ascontiguousarray(initial_bearing, dtype=np.float64)
    thickness_array = np.empty(num_elements, dtype=np.float64)

    # Make memory view of the numpy arrays
    cdef cnp.float64_t[::1] r = radial_distance
    cdef cnp.float64_t[::1] theta = initial_bearing
    cdef cnp.float64_t[::1] ejecta_thickness = thickness_array

    bind_ejecta_ray_intensity(&r[0], &theta[0], num_elements, crater_diameter, ejrim, ejecta_truncation, &ejecta_thickness[0])

    return ejecta_thickness  




import unittest
import numpy as np
from cratermaker.utils.montecarlo import *
from scipy.stats import chisquare, ks_2samp

class TestMonteCarlo(unittest.TestCase):

    def test_get_random_location(self):
        """Test get_random_location with default parameters."""
        # Do three trials. If all three fail, the test fails.
        ntrials = 3
        trial_results = []
        for _ in range(ntrials):
            size = 1000
            points = get_random_location(size=size)

            lons = points['lon']
            lats = points['lat']
            self.assertIsInstance(lons[0], np.float64)
            self.assertIsInstance(lats[0], np.float64)
            self.assertTrue(np.all(-180.0 <= lons) and np.all(lons <= 180.0), msg=f"Longitude range: {np.min(lons)} to {np.max(lons)}")
            self.assertTrue(np.all(-90.0 <= lats) and np.all(lats <= 90.0), msg=f"Latitude range: {np.min(lats)} to {np.max(lats)}")
            
            # Test Longitude Uniformity
            
            # Set a significance level for your test, commonly 0.05
            alpha = 0.05
        
            bins = 20 
            observed_counts, bins_lon = np.histogram(lons, bins=bins, range=(-180, 180))
            expected_count_lon = size // bins
            
            # Perform the chi-square test
            _, p_value = chisquare(f_obs=observed_counts, f_exp=[expected_count_lon]*bins)

            # Assert that the p-value is greater than the significance level
            result = p_value
            
            sin_lats = np.sin(np.deg2rad(lats))
            observed_counts, _ = np.histogram(sin_lats, bins=bins, range=(-1, 1))
            expected_count_lat = size // bins
            
            # Perform the chi-square test
            _, p_value = chisquare(f_obs=observed_counts, f_exp=[expected_count_lat]*bins)

            # Assert that the p-value is greater than the significance level
            result = result and p_value > alpha
            trial_results.append(result)
        self.assertTrue(any(trial_results)) 
        return
    

    def test_get_random_impact_angle(self):
        """Test get_random_impact_angle function."""
        size = 10000
        angles = get_random_impact_angle(size=size)

        # Check type and morphology
        self.assertIsInstance(angles, np.ndarray)
        self.assertEqual(angles.shape, (size,))

        # Check range (0 to np.pi/2 radians)
        self.assertTrue(np.all(angles >= 0) and np.all(angles <= 90.0))

        # Check distribution
        # Generate expected distribution centered around 45 degrees
        expected_angles = np.arcsin(np.sqrt(np.linspace(0, 1, size)))
        _, p_value = ks_2samp(angles, np.rad2deg(expected_angles))

        # Check if the p-value is greater than a chosen significance level (0.05)
        self.assertGreater(p_value, 0.05)


    def test_get_random_size(self):
        # Generate power law CDF and test that it results in a Poisson-like distribution of binned counts
        nbins = 10
        size = 10000
        num_realizations = 1000
        
        Dhi = 100.0
        p = 3.0
        C =  Dhi**p 
        Dlo = (size/C)**(-1.0/p) 
        diameters = np.exp(np.linspace(np.log(Dlo), np.log(100*Dhi), nbins))
        model_cdf = C*diameters**-p

        # Generate multiple populations of sizes and bin the results from each population
        all_counts = []
        for _ in range(num_realizations):
            random_sizes = get_random_size(diameters, model_cdf, mu=size)
            counts, _ = np.histogram(random_sizes, bins=diameters)
            cumulative_counts = np.cumsum(counts[::-1])[::-1]
            all_counts.append(cumulative_counts)
       
        # Assuming 'all_counts' is a list of arrays, each array being the counts for one realization
        observed_counts = np.stack(all_counts)  # Stack all count arrays

        # Calculate the mean and standard deviation of observed counts for each bin
        observed_means = np.mean(observed_counts, axis=0)
        observed_std_devs = np.std(observed_counts, axis=0)

        # Calculate the expected standard deviation for each bin from the Poisson distribution
        expected_counts = model_cdf 
        expected_std_devs = np.sqrt(expected_counts)

        # Compare observed and expected standard deviations
        for i in range(len(observed_means)):
            if observed_means[i] == 0:
                continue
            self.assertAlmostEqual(observed_means[i], expected_counts[i], delta=4*expected_std_devs[i])
            
        # test with invalid diameter shape or size
        diameters = np.array([[100., 56.], [32., 18.]])  # 2D array
        cdf = np.array([1., 0.8, 0.6, 0.4])
        with self.assertRaises(ValueError):
            get_random_size(diameters, cdf)
        diameters = np.array([])  # Empty array
        cdf = np.array([])
        with self.assertRaises(ValueError):
            get_random_size(diameters, cdf) 
        diameters = np.array([1.0])  # Single element array
        cdf = np.array([1.0])
        with self.assertRaises(ValueError):
            get_random_size(diameters, cdf)
        diameters = 1.0  # Scalar
        cdf = 1.0
        with self.assertRaises(ValueError):
            get_random_size(diameters, cdf)                          
        diameters = np.array([100., 56., 32., 18.])
        cdf = np.array([0.4, 0.6, 0.8])
        with self.assertRaises(ValueError):
            get_random_size(diameters, cdf)
        
        # test the case when the cdf is not monotonically decreasing with increasing diameter 
        diameters = np.array([100., 56., 32., 18.])
        cdf = np.array([1., 0.8, 0.6, 0.4])
        with self.assertRaises(ValueError):
            get_random_size(diameters, cdf)
            
        # test the case when both size and mu are requested
        diameters = np.logspace(0,3,nbins)
        model_cdf = diameters**-3
        with self.assertRaises(ValueError):
            get_random_size(diameters, model_cdf, size=100, mu=100)
         
        # test the case when negative values are requested
        diameters = np.logspace(0,3,nbins)
        model_cdf = diameters**-3
        with self.assertRaises(ValueError):
            get_random_size(diameters, model_cdf, size=-100)
        with self.assertRaises(ValueError):
            get_random_size(diameters, model_cdf, mu=-100)
        with self.assertRaises(ValueError):
            get_random_size(diameters, -model_cdf)
        with self.assertRaises(ValueError):
            get_random_size(-diameters, model_cdf)

        return
        

if __name__ == '__main__':
    unittest.main()
